"""
Windows 檔案關聯 — 「以 Imervue 開啟」右鍵選單
Windows file association — "Open with Imervue" context menu.
"""
from __future__ import annotations

import logging
import os
import sys
from pathlib import Path

from Imervue.system.app_paths import is_frozen, app_dir, icon_path as _app_icon_path

logger = logging.getLogger("Imervue.file_assoc")

# 支援關聯的副檔名
ASSOC_EXTENSIONS = [
    ".png", ".jpg", ".jpeg", ".bmp", ".tiff", ".tif",
    ".webp", ".gif", ".apng",
    ".cr2", ".nef", ".arw", ".dng", ".raf", ".orf",
]

_APP_ID = "Imervue.ImageViewer"
_SHELL_LABEL = "Open with Imervue"


def _build_command() -> str:
    """建立用於檔案關聯的啟動命令。

    * 凍結環境：直接用 exe 路徑。
    * 開發環境：用 python + 入口腳本 / -m Imervue。
    """
    if is_frozen():
        return f'"{sys.executable}" "%1"'

    python_exe = sys.executable
    pkg_dir = app_dir() / "Imervue"
    main_file = pkg_dir / "__main__.py"
    if main_file.exists():
        return f'"{python_exe}" "{main_file}" "%1"'
    return f'"{python_exe}" -m Imervue "%1"'


def register_file_association() -> tuple[bool, str]:
    """在 Windows 登錄檔中註冊檔案關聯。需要管理員權限。

    Returns:
        (success, message)
    """
    if sys.platform != "win32":
        return False, "Only supported on Windows"

    try:
        import winreg
    except ImportError:
        return False, "winreg not available"

    command = _build_command()

    try:
        # 1. 建立 Application 登錄
        app_key_path = rf"Software\Classes\{_APP_ID}"
        with winreg.CreateKey(winreg.HKEY_CURRENT_USER, app_key_path) as key:
            winreg.SetValueEx(key, "", 0, winreg.REG_SZ, "Imervue Image Viewer")

        # shell\open\command
        cmd_path = rf"{app_key_path}\shell\open\command"
        with winreg.CreateKey(winreg.HKEY_CURRENT_USER, cmd_path) as key:
            winreg.SetValueEx(key, "", 0, winreg.REG_SZ, command)

        # 2. 為每種副檔名加入「以 Imervue 開啟」右鍵選單
        for ext in ASSOC_EXTENSIONS:
            _register_ext_context_menu(ext, command)

        logger.info("File association registered successfully")
        return True, "OK"
    except PermissionError:
        return False, "need_admin"
    except Exception as e:
        logger.error(f"Registration failed: {e}")
        return False, str(e)


def _register_ext_context_menu(ext: str, command: str):
    """為指定副檔名加入右鍵 shell 選單"""
    import winreg

    # HKCU\Software\Classes\.ext\shell\Imervue\command
    shell_path = rf"Software\Classes\SystemFileAssociations\{ext}\shell\Imervue"
    with winreg.CreateKey(winreg.HKEY_CURRENT_USER, shell_path) as key:
        winreg.SetValueEx(key, "", 0, winreg.REG_SZ, _SHELL_LABEL)
        # Icon
        ico = _app_icon_path()
        if ico.exists():
            winreg.SetValueEx(key, "Icon", 0, winreg.REG_SZ, str(ico))

    cmd_path = rf"{shell_path}\command"
    with winreg.CreateKey(winreg.HKEY_CURRENT_USER, cmd_path) as key:
        winreg.SetValueEx(key, "", 0, winreg.REG_SZ, command)


def unregister_file_association() -> tuple[bool, str]:
    """移除已註冊的檔案關聯。

    Returns:
        (success, message)
    """
    if sys.platform != "win32":
        return False, "Only supported on Windows"

    try:
        import winreg
    except ImportError:
        return False, "winreg not available"

    try:
        # 移除 Application 登錄
        _delete_key_tree(winreg.HKEY_CURRENT_USER, rf"Software\Classes\{_APP_ID}")

        # 移除各副檔名的右鍵選單
        for ext in ASSOC_EXTENSIONS:
            shell_path = rf"Software\Classes\SystemFileAssociations\{ext}\shell\Imervue"
            _delete_key_tree(winreg.HKEY_CURRENT_USER, shell_path)

        logger.info("File association removed successfully")
        return True, "OK"
    except PermissionError:
        return False, "need_admin"
    except Exception as e:
        logger.error(f"Unregistration failed: {e}")
        return False, str(e)


def _delete_key_tree(hive, path: str):
    """遞迴刪除登錄機碼"""
    import winreg
    try:
        with winreg.OpenKey(hive, path, 0, winreg.KEY_ALL_ACCESS) as key:
            # 先刪除所有子機碼
            while True:
                try:
                    subkey_name = winreg.EnumKey(key, 0)
                    _delete_key_tree(hive, rf"{path}\{subkey_name}")
                except OSError:
                    break
        winreg.DeleteKey(hive, path)
    except FileNotFoundError:
        pass  # 機碼不存在



"""
縮圖磁碟快取
Persistent thumbnail disk cache — avoids re-decoding images on every folder open.
Uses .npy format for fast I/O (raw numpy array, no encode/decode overhead).
Cache key = md5(absolute_path | mtime_ns | file_size | thumbnail_size).
"""
from __future__ import annotations

import hashlib
import logging
import os
import sys
from pathlib import Path

import numpy as np

logger = logging.getLogger("Imervue.thumbnail_cache")


def _get_cache_dir() -> Path:
    if sys.platform == "win32":
        base = Path(os.environ.get("LOCALAPPDATA", str(Path.home())))
        return base / "Imervue" / "cache" / "thumbnails"
    return Path.home() / ".cache" / "imervue" / "thumbnails"


class ThumbnailDiskCache:

    def __init__(self):
        self._dir = _get_cache_dir()
        try:
            self._dir.mkdir(parents=True, exist_ok=True)
        except OSError as e:
            logger.warning(f"Cannot create cache dir {self._dir}: {e}")

    # --------------------------------------------------

    @staticmethod
    def _key(path: str, size: int) -> str:
        try:
            st = Path(path).stat()
            raw = f"{path}|{st.st_mtime_ns}|{st.st_size}|{size}"
        except OSError:
            return ""
        return hashlib.md5(raw.encode()).hexdigest()

    # --------------------------------------------------

    def get(self, path: str, size: int) -> np.ndarray | None:
        """嘗試從磁碟讀取快取的縮圖，失效或不存在時回傳 None"""
        key = self._key(path, size)
        if not key:
            return None
        cache_file = self._dir / f"{key}.npy"
        if cache_file.exists():
            try:
                return np.load(cache_file)
            except Exception:
                cache_file.unlink(missing_ok=True)
        return None

    def put(self, path: str, size: int, img_data: np.ndarray) -> None:
        """將縮圖寫入磁碟快取"""
        key = self._key(path, size)
        if not key:
            return
        cache_file = self._dir / f"{key}.npy"
        try:
            np.save(cache_file, img_data)
        except Exception as e:
            logger.debug(f"Failed to write thumbnail cache: {e}")


# 模組層級單例
thumbnail_disk_cache = ThumbnailDiskCache()



"""Tests for SVG loading support."""
import numpy as np
import pytest


class TestSVGLoader:
    def test_svg_in_supported_exts(self):
        from Imervue.gpu_image_view.images.image_loader import _SUPPORTED_EXTS
        assert ".svg" in _SUPPORTED_EXTS

    def test_load_svg_file(self, tmp_path):
        """A minimal SVG should load as RGBA numpy array."""
        svg_path = tmp_path / "test.svg"
        svg_path.write_text(
            '<svg xmlns="http://www.w3.org/2000/svg" width="64" height="64">'
            '<rect width="64" height="64" fill="red"/>'
            '</svg>'
        )
        from Imervue.gpu_image_view.images.image_loader import load_image_file
        result = load_image_file(str(svg_path))
        assert result.ndim == 3
        assert result.shape[2] == 4
        assert result.shape[0] == 64
        assert result.shape[1] == 64
        # Red rect → R channel should be dominant
        assert result[:, :, 0].mean() > 200

    def test_scan_finds_svg(self, tmp_path):
        """_scan_images should find .svg files."""
        svg_path = tmp_path / "icon.svg"
        svg_path.write_text(
            '<svg xmlns="http://www.w3.org/2000/svg" width="10" height="10"/>'
        )
        from Imervue.gpu_image_view.images.image_loader import _scan_images
        result = _scan_images(str(tmp_path))
        assert any(p.endswith(".svg") for p in result)



import math

import numpy as np
from PIL import Image


class DeepZoomImage:
    def __init__(self, image_array):
        self.levels = []
        self.build_pyramid(image_array)

    def build_pyramid(self, image):
        current = image
        self.levels.append(current)

        while min(current.shape[:2]) > 512:
            h, w = current.shape[:2]
            new_h, new_w = h // 2, w // 2
            # 使用 PIL LANCZOS 進行高品質降採樣，避免鋸齒
            channels = current.shape[2] if current.ndim == 3 else 1
            if channels == 4:
                pil_img = Image.fromarray(current, "RGBA")
            elif channels == 3:
                pil_img = Image.fromarray(current, "RGB")
            else:
                pil_img = Image.fromarray(current)
            pil_img = pil_img.resize((new_w, new_h), Image.Resampling.LANCZOS)
            current = np.array(pil_img)
            self.levels.append(current)

    def get_level(self, zoom):
        # zoom 1.0 = full res
        level = int(max(0, min(len(self.levels)-1, -math.log2(zoom))))
        return level, self.levels[level]



"""Tests for ImageModel."""
from Imervue.gpu_image_view.images.image_model import ImageModel


class TestImageModel:
    def test_init_empty(self):
        m = ImageModel()
        assert m.images == []
        assert m.deleted_history == []

    def test_set_images(self):
        m = ImageModel()
        paths = ["/a.png", "/b.png", "/c.png"]
        m.set_images(paths)
        assert m.images == paths
        # Should be a copy, not same reference
        assert m.images is not paths

    def test_delete_single(self):
        m = ImageModel()
        m.set_images(["/a.png", "/b.png", "/c.png"])
        deleted = m.delete_images(["/b.png"])
        assert m.images == ["/a.png", "/c.png"]
        assert len(deleted) == 1
        assert deleted[0] == ("/b.png", 1)

    def test_delete_multiple(self):
        m = ImageModel()
        m.set_images(["/a.png", "/b.png", "/c.png", "/d.png"])
        deleted = m.delete_images(["/a.png", "/c.png"])
        assert m.images == ["/b.png", "/d.png"]
        assert len(deleted) == 2

    def test_delete_nonexistent(self):
        m = ImageModel()
        m.set_images(["/a.png"])
        deleted = m.delete_images(["/z.png"])
        assert m.images == ["/a.png"]
        assert deleted == []

    def test_restore_images(self):
        m = ImageModel()
        m.set_images(["/a.png", "/b.png", "/c.png"])
        deleted = m.delete_images(["/b.png"])
        assert m.images == ["/a.png", "/c.png"]
        m.restore_images(deleted)
        assert m.images == ["/a.png", "/b.png", "/c.png"]

    def test_restore_multiple_preserves_order(self):
        m = ImageModel()
        m.set_images(["/a.png", "/b.png", "/c.png", "/d.png"])
        deleted = m.delete_images(["/a.png", "/c.png"])
        m.restore_images(deleted)
        assert m.images == ["/a.png", "/b.png", "/c.png", "/d.png"]



korean_word_dict = {
    "main_window_current_filename_format": "현재 파일 이름: {name}",
    # File menu
    "main_window_open_image": "파일 열기",
    "main_window_current_filename": "현재 파일 이름:",
    "main_window_current_file": "파일",
    "main_window_open_folder": "폴더 열기",
    "main_window_exit": "종료",
    "main_window_tile_size": "썸네일 크기",
    "main_window_select_folder": "폴더 선택",
    "main_window_select_image": "이미지 선택",
    "main_window_current_folder_format": "현재 폴더: {path}",
    "main_window_remove_undo_stack": "임시 되돌리기 파일 삭제",
    # Language Menubar
    "menu_bar_language": "언어",
    "language_menu_bar_please_restart_messagebox": "다시 시작해 주세요",
    "language_menu_bar_english": "영어",
    "language_menu_bar_traditional_chinese": "번체 중국어",
    "language_menu_bar_chinese": "간체 중국어",
    "language_menu_bar_koren": "한국어",
    "language_menu_bar_japanese": "일본어",
    # Tip menu
    "main_window_tip_menu": "사용 설명",
    "mouse_control_middle_tip": "마우스 휠 버튼으로 이동할 수 있습니다",
    "main_window_mouse_tip_menu": "마우스 조작",
    "main_window_keyboard_tip_menu": "키보드 조작",
    "mouse_control_left_tip": "마우스 왼쪽 버튼으로 이미지를 클릭하면 Deep zoom 모드로 들어갑니다",
    "mouse_control_multi_select_tip": "마우스 왼쪽 버튼을 누른 상태에서 썸네일 모드에서 여러 이미지를 선택할 수 있습니다",
    "keyboard_control_esc_tip": "Deep zoom 모드에서는 ESC 키로 종료할 수 있습니다",
    "keyboard_control_tile_arrow_up_tip": "썸네일 모드에서는 방향키 위로 위쪽으로 이동할 수 있습니다",
    "keyboard_control_tile_arrow_down_tip": "썸네일 모드에서는 방향키 아래로 아래쪽으로 이동할 수 있습니다",
    "keyboard_control_tile_arrow_left_tip": "썸네일 모드에서는 방향키 왼쪽으로 왼쪽으로 이동할 수 있습니다",
    "keyboard_control_tile_arrow_right_tip": "썸네일 모드에서는 방향키 오른쪽으로 오른쪽으로 이동할 수 있습니다",
    "keyboard_control_delete_tip": "Delete 키로 이미지를 휴지통으로 이동할 수 있습니다 (썸네일 및 Deep zoom 모드)",
    "keyboard_r_tip": "Deep zoom 모드에서 R / Shift+R로 시계/반시계 방향 회전",
    "keyboard_home_tip": "Home 키로 확대/축소 및 위치를 기본값으로 초기화",
    "keyboard_e_tip": "E 키로 이미지 편집기 열기",
    "keyboard_undo_tip": "마지막 작업을 실행 취소",
    "keyboard_redo_tip": "실행 취소한 작업을 다시 실행",
    "keyboard_arrow_lr_tip": "Deep zoom 모드에서 좌/우 키로 이미지 전환",
    "keyboard_arrow_tile_tip": "썸네일 모드에서 방향키로 이동",
    "tip_mouse_middle": "가운데 버튼",
    "tip_mouse_left_click": "왼쪽 클릭",
    "tip_mouse_left_drag": "왼쪽 드래그",
    "tip_mouse_right_desc": "컨텍스트 메뉴를 열어 다양한 작업 수행",
    "tip_mouse_scroll_desc": "스크롤로 확대/축소 또는 썸네일 스크롤",
    "tip_close": "닫기",
    "tip_show_shortcuts": "키보드 및 마우스 단축키",
    "keyboard_f_tip": "F 키로 전체화면 전환, Esc로 종료",
    "keyboard_ctrl_c_tip": "Ctrl+C로 현재 이미지를 클립보드에 복사",
    "keyboard_rating_tip": "1~5 키로 현재 이미지 평가; 다시 누르면 해제",
    "keyboard_search_tip": "/ 또는 Ctrl+F로 이미지 이름 검색",
    "keyboard_favorite_tip": "0 키로 현재 이미지 즐겨찾기 전환",
    "keyboard_slideshow_tip": "S 키로 슬라이드쇼 시작/중지",
    # 평가 상태
    "rating_set": "평가: {star}",
    "rating_cleared": "평가가 해제되었습니다",
    # 검색 다이얼로그
    "search_dialog_title": "이미지 검색",
    "search_dialog_placeholder": "검색어를 입력하세요...",
    "search_open": "열기",
    # Right click menu
    "right_click_menu_go_to_parent_folder": "상위 폴더로 이동",
    "right_click_menu_next_image": "다음 이미지",
    "right_click_menu_previous_image": "이전 이미지",
    "right_click_menu_delete_current": "현재 이미지 삭제",
    "right_click_menu_delete_selected": "선택한 이미지 삭제",
    "right_click_menu_image_info": "이미지 정보",
    # Image info
    "image_info_filename": "파일명: {info}\n",
    "image_info_fullpath": "전체 경로: {full_path}\n",
    "image_info_image_size": "크기: {width} x {height}\n",
    "image_info_file_size": "파일 크기: {file_size_mb} MB\n",
    "image_info_file_created_time": "생성 시간: {created_time}\n",
    "image_info_file_modified_time": "수정 시간: {modified_time}\n",
    "image_info_messagebox_title": "이미지 정보",
    # Image info exif
    "image_info_exif_datatime_original": "촬영 시간: {DateTimeOriginal}\n",
    "image_info_exif_camera_model": "카메라: {Make} {Model}\n",
    "image_info_exif_camera_lens_model": "렌즈: {LensModel}\n",
    "image_info_exif_camera_focal_length": "초점 거리: {FocalLength}\n",
    "image_info_exif_camera_fnumber": "조리개: {FNumber}\n",
    "image_info_exif_exposure_time": "셔터: {ExposureTime}\n",
    "image_info_exif_iso": "ISO: {ISOSpeedRatings}",
    # Recent menu
    "recent_menu_title": "최근 열린 메뉴",
    # Plugin
    "plugin_menu_title": "플러그인",
    "plugin_menu_loaded": "로드된 플러그인",
    "plugin_menu_no_plugins": "로드된 플러그인이 없습니다",
    "plugin_menu_reload": "플러그인 다시 로드",
    "plugin_menu_open_folder": "플러그인 폴더 열기",
    "plugin_info_name": "이름: {name}",
    "plugin_info_version": "버전: {version}",
    "plugin_info_author": "작성자: {author}",
    "plugin_info_description": "설명: {description}",
    "plugin_info_languages": "지원 언어: {languages}",
    "plugin_menu_manage": "플러그인 관리",
    "plugin_manage_count": "{count}개의 플러그인이 로드됨",
    "plugin_manage_select_hint": "플러그인을 선택하여 세부 정보를 확인하세요",
    "plugin_reload_confirm": "모든 플러그인을 다시 로드하시겠습니까? 현재 플러그인을 언로드하고 다시 검색합니다.",
    "plugin_reload_done": "{count}개의 플러그인을 다시 로드했습니다",
    "plugin_menu_download": "플러그인 다운로드",
    # Plugin Downloader
    "plugin_dl_title": "플러그인 다운로드",
    "plugin_dl_refresh": "새로 고침",
    "plugin_dl_col_name": "플러그인",
    "plugin_dl_col_status": "상태",
    "plugin_dl_download": "다운로드",
    "plugin_dl_reinstall": "재설치",
    "plugin_dl_delete": "삭제",
    "plugin_dl_close": "닫기",
    "plugin_dl_installed": "설치됨",
    "plugin_dl_not_installed": "미설치",
    "plugin_dl_status_ready": "준비 완료",
    "plugin_dl_status_fetching": "플러그인 목록을 가져오는 중...",
    "plugin_dl_status_found": "{count}개의 플러그인을 찾았습니다",
    "plugin_dl_status_error": "오류: {error}",
    "plugin_dl_status_downloading": "{name} 다운로드 중...",
    "plugin_dl_status_done": "{name} 다운로드 완료!",
    "plugin_dl_status_deleted": "{name}이(가) 삭제되었습니다.",
    "plugin_dl_restart_hint": "플러그인 '{name}'이(가) 다운로드되었습니다. Imervue를 다시 시작하여 활성화하세요.",
    "plugin_dl_delete_confirm_title": "플러그인 삭제",
    "plugin_dl_delete_confirm": "플러그인 '{name}'을(를) 삭제하시겠습니까?",
    # 우클릭 메뉴
    "right_click_show_in_explorer": "탐색기에서 표시",
    "right_click_copy_path": "경로 복사",
    "right_click_copy_image": "이미지 복사",
    "right_click_rotate_cw": "시계 방향 회전",
    "right_click_rotate_ccw": "반시계 방향 회전",
    "right_click_set_wallpaper": "배경화면으로 설정",
    "right_click_compare": "이미지 비교",
    "right_click_slideshow": "슬라이드쇼",
    "right_click_slideshow_stop": "슬라이드쇼 중지",
    # 정렬 메뉴
    "sort_menu_title": "정렬",
    "sort_by_name": "이름순",
    "sort_by_modified": "수정일순",
    "sort_by_created": "생성일순",
    "sort_by_size": "파일 크기순",
    "sort_by_resolution": "해상도순",
    "sort_ascending": "오름차순",
    "sort_descending": "내림차순",
    # 필터 메뉴
    "filter_menu_title": "필터",
    "filter_by_extension": "확장자별",
    "filter_by_rating": "평가별",
    "filter_ext_all": "전체",
    "filter_ext_jpg": "JPEG",
    "filter_ext_png": "PNG",
    "filter_ext_bmp": "BMP",
    "filter_ext_tiff": "TIFF",
    "filter_ext_raw": "RAW",
    "filter_rating_all": "전체",
    "filter_rating_favorited": "즐겨찾기",
    "filter_clear": "필터 해제",
    # 즐겨찾기
    "favorite_added": "\u2764 즐겨찾기에 추가됨",
    "favorite_removed": "즐겨찾기 해제됨",
    # 비교
    "compare_title": "이미지 비교",
    "compare_2": "2장 비교",
    "compare_4": "4장 비교",
    # EXIF 사이드바
    "exif_filename": "파일",
    "exif_filesize": "크기",
    "exif_created": "생성일시",
    "exif_modified": "수정일시",
    "exif_taken": "촬영일시",
    "exif_make": "제조사",
    "exif_model": "모델",
    "exif_lens": "렌즈",
    "exif_focal": "초점 거리",
    "exif_aperture": "조리개",
    "exif_shutter": "셔터 속도",
    "exif_iso": "ISO",
    "exif_ev": "노출 보정",
    "exif_wb": "화이트밸런스",
    "exif_flash": "플래시",
    "exif_resolution": "해상도",
    "exif_no_data": "EXIF 데이터 없음",
    # 슬라이드쇼
    "slideshow_title": "슬라이드쇼",
    "slideshow_interval": "간격 (초):",
    "slideshow_fade": "페이드 전환",
    "slideshow_start": "시작",
    "slideshow_cancel": "취소",
    # 상태 표시줄
    "status_loading_progress": "로딩 중 {current}/{total}...",
    "status_loading_image": "이미지 로딩 중...",
    "status_ready": "준비 완료",
    # Toast
    "toast_path_copied": "경로가 복사되었습니다",
    # 우클릭 추가
    "right_click_edit_image": "이미지 편집",
    "right_click_batch": "일괄 작업",
    "batch_rename_title": "일괄 이름 변경",
    "batch_rename_hint": "템플릿: {name} = 원래 이름, {n} = 일련번호, {ext} = 확장자",
    "batch_rename_start": "시작 번호:",
    "batch_rename_apply": "이름 변경",
    "batch_rename_cancel": "취소",
    "batch_move_title": "이동 / 복사",
    "batch_move_mode": "모드",
    "batch_move_move": "이동",
    "batch_move_copy": "복사",
    "batch_move_apply": "적용",
    "batch_move_cancel": "취소",
    "batch_rotate_cw": "모두 시계 방향 회전",
    "batch_rotate_ccw": "모두 반시계 방향 회전",
    # 이미지 편집기
    "editor_title": "이미지 편집기",
    "editor_brightness": "밝기",
    "editor_contrast": "대비",
    "editor_saturation": "채도",
    "editor_apply_crop": "자르기 적용",
    "editor_rotate_cw": "시계 방향 회전",
    "editor_rotate_ccw": "반시계 방향 회전",
    "editor_reset": "초기화",
    "editor_save_as": "다른 이름으로 저장...",
    "editor_save": "저장",
    "editor_saved": "저장되었습니다!",
    # EXIF 편집기
    "exif_edit_button": "EXIF 편집",
    "exif_editor_title": "EXIF 메타데이터 편집",
    "exif_editor_no_piexif": "piexif 패키지가 필요합니다.\n설치: pip install piexif",
    "exif_editor_fields": "메타데이터 필드",
    "exif_editor_save": "저장",
    "exif_editor_cancel": "취소",
    "exif_editor_close": "닫기",
    "exif_editor_saved": "EXIF가 저장되었습니다!",
    "exif_edit_description": "설명",
    "exif_edit_artist": "아티스트",
    "exif_edit_copyright": "저작권",
    "exif_edit_make": "제조사",
    "exif_edit_model": "모델",
    "exif_edit_user_comment": "사용자 코멘트",
    # 내보내기 대화상자
    "export_title": "이미지 내보내기",
    "export_format": "형식:",
    "export_quality": "품질:",
    "export_browse": "찾아보기...",
    "export_save": "저장",
    "export_cancel": "취소",
    "export_saved": "내보내기 완료!",
    # 히스토그램
    "histogram_toggle": "H 키로 RGB 히스토그램 전환",
    # 무손실 회전
    "lossless_rotate_cw": "무손실 시계 방향 회전",
    "lossless_rotate_ccw": "무손실 반시계 방향 회전",
    "lossless_rotate_done": "무손실 회전이 저장되었습니다!",
    "lossless_rotate_fail": "무손실 회전 실패",
    # 우클릭 메뉴 추가
    "right_click_export": "내보내기 / 다른 이름으로 저장...",
    "right_click_lossless_rotate": "무손실 회전",
    # 애니메이션 재생
    "anim_play": "재생",
    "anim_pause": "일시 정지",
    "anim_frame_prev": "이전 프레임",
    "anim_frame_next": "다음 프레임",
    "anim_frame_indicator": "프레임 {current}/{total}",
    "anim_speed": "속도: {speed}x",
    # 파일 연결
    "file_assoc_menu": "파일 연결",
    "file_assoc_register": "'Imervue로 열기' 등록",
    "file_assoc_unregister": "파일 연결 제거",
    "file_assoc_done": "파일 연결이 등록되었습니다!",
    "file_assoc_removed": "파일 연결이 제거되었습니다!",
    "file_assoc_need_admin": "관리자 권한 필요",
    # 다중 창
    "menu_new_window": "새 창",
    # 맞춤 모드
    "fit_width": "너비 맞춤",
    "fit_height": "높이 맞춤",
    "keyboard_w_tip": "W로 너비 맞춤, Shift+W로 높이 맞춤",
    # 추가 키보드 힌트
    "keyboard_h_tip": "딥줌 모드에서 H 키로 RGB 히스토그램 전환",
    "keyboard_ctrl_v_tip": "Ctrl+V로 클립보드에서 이미지 붙여넣기",
    "keyboard_anim_play_tip": "스페이스 키로 애니메이션 이미지(GIF/APNG/WebP) 재생/일시 정지",
    "keyboard_anim_frame_tip": "< / >로 애니메이션 프레임 이동",
    "keyboard_anim_speed_tip": "[ / ]로 애니메이션 속도 조절",
    # 북마크
    "bookmark_title": "북마크",
    "bookmark_open": "열기",
    "bookmark_remove": "삭제",
    "bookmark_clear": "모두 삭제",
    "bookmark_close": "닫기",
    "bookmark_count": "{count}개의 북마크된 이미지",
    "bookmark_added": "\u2605 북마크 추가됨",
    "bookmark_removed": "북마크 해제됨",
    "bookmark_add_current": "북마크에 추가",
    "bookmark_remove_current": "북마크 해제",
    "bookmark_clear_confirm_title": "북마크 삭제",
    "bookmark_clear_confirm": "모든 북마크를 삭제하시겠습니까?",
    "keyboard_b_tip": "B 키로 현재 이미지 북마크 전환",
    # SVG
    "filter_ext_svg": "SVG",
    # 태그 & 앨범
    "tag_album_title": "태그 & 앨범",
    "tag_tab_tags": "태그",
    "tag_tab_albums": "앨범",
    "tag_close": "닫기",
    "tag_count": "{count}개의 태그",
    "tag_create": "새 태그",
    "tag_rename": "이름 변경",
    "tag_delete": "삭제",
    "tag_images": "태그의 이미지:",
    "tag_remove_image": "태그에서 제거",
    "tag_filter": "뷰어에서 보기",
    "tag_create_prompt": "태그 이름:",
    "tag_rename_prompt": "새 이름:",
    "tag_delete_confirm_title": "태그 삭제",
    "tag_delete_confirm": "태그 '{name}'을(를) 삭제하시겠습니까?",
    "tag_menu_title": "태그",
    "tag_create_and_add": "+ 새 태그...",
    "album_count": "{count}개의 앨범",
    "album_create": "새 앨범",
    "album_rename": "이름 변경",
    "album_delete": "삭제",
    "album_images": "앨범의 이미지:",
    "album_remove_image": "앨범에서 제거",
    "album_view": "앨범 보기",
    "album_create_prompt": "앨범 이름:",
    "album_rename_prompt": "새 이름:",
    "album_delete_confirm_title": "앨범 삭제",
    "album_delete_confirm": "앨범 '{name}'을(를) 삭제하시겠습니까?",
    "album_menu_title": "앨범",
    "album_create_and_add": "+ 새 앨범...",
    "batch_tag_title": "태그에 추가",
    "batch_album_title": "앨범에 추가",
    "filter_by_tag": "태그별",
    "filter_by_album": "앨범별",
    "filter_no_tags": "(태그 없음)",
    "filter_no_albums": "(앨범 없음)",
    "keyboard_t_tip": "T 키로 태그 & 앨범 관리 열기",
    # 일괄 내보내기
    "batch_export_title": "일괄 내보내기",
    "batch_export_count": "{count}개의 이미지 선택됨",
    "batch_export_resize": "크기 조정",
    "batch_export_max_width": "최대 너비:",
    "batch_export_max_height": "최대 높이:",
    "batch_export_start": "내보내기",
    "batch_export_done": "{success}/{total}개의 이미지를 내보냈습니다",
    # GIF / 동영상
    "gif_video_title": "GIF / 동영상 만들기",
    "gif_video_order": "드래그하여 순서 변경 (위 = 첫 프레임):",
    "gif_video_move_up": "위로",
    "gif_video_move_down": "아래로",
    "gif_video_settings": "설정",
    "gif_video_fps": "FPS:",
    "gif_video_width": "너비:",
    "gif_video_height": "높이:",
    "gif_video_auto": "자동",
    "gif_video_loop": "무한 반복",
    "gif_video_create": "만들기",
    "gif_video_save": "다른 이름으로 저장",
    "gif_video_done": "생성 완료: {path}",
}



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from Imervue.Imervue_main_window import ImervueMainWindow

from Imervue.menu.plugin_menu import build_plugin_menu
from Imervue.multi_language.language_wrapper import language_wrapper
from Imervue.plugin.pip_installer import register_translations as _register_pip_translations
from Imervue.plugin.plugin_manager import PluginManager

logger = logging.getLogger("Imervue.integration")


def _init_plugin_system_example(main_window: ImervueMainWindow) -> None:
    """Initialize the plugin system and load all discovered plugins.

    Called from ImervueMainWindow.__init__() after the UI is fully built.
    """
    # 註冊 pip 安裝對話框的翻譯（供所有插件共用）
    logger.info("Initializing plugin system")
    _register_pip_translations()

    manager = PluginManager(main_window)
    logger.info("Discovering and loading plugins...")
    manager.discover_and_load()
    logger.info("Plugin loading complete, %d plugin(s) loaded", len(manager.plugins))

    # Store on main_window so other parts of the app can access it
    main_window.plugin_manager = manager

    # If plugins registered new languages, append them to the existing language menu
    if language_wrapper.plugin_languages and hasattr(main_window, "language_menu"):
        from Imervue.menu.language_menu import set_language
        from PySide6.QtGui import QAction
        main_window.language_menu.addSeparator()
        for lang_code, display_name in language_wrapper.plugin_languages.items():
            action = QAction(display_name, main_window.language_menu)
            action.triggered.connect(
                lambda _, code=lang_code: set_language(code, main_window)
            )
            main_window.language_menu.addAction(action)

    # Let plugins add their own menus
    manager.dispatch_build_menu_bar(main_window.menuBar())

    # Build the plugin management menu (after plugins are loaded so it can list them)
    build_plugin_menu(main_window)



import sys
from pathlib import Path

from PySide6.QtCore import Qt, QTimer, QDir, QFileSystemWatcher
from PySide6.QtGui import QIcon
from PySide6.QtWidgets import (
    QApplication, QMainWindow, QTreeView, QFileSystemModel, QSplitter, QSizePolicy,
    QStatusBar, QProgressBar, QHeaderView,
)
from PySide6.QtWidgets import QLabel, QVBoxLayout, QWidget

from Imervue.system.app_paths import icon_path as _app_icon_path
from Imervue.gpu_image_view.actions.delete import commit_pending_deletions
from Imervue.gpu_image_view.images.image_loader import open_path
from Imervue.gui.exif_sidebar import ExifSidebar
from Imervue.gui.toast import ToastManager
from Imervue.integration_guide import _init_plugin_system_example
from Imervue.menu.file_menu import build_file_menu
from Imervue.menu.filter_menu import build_filter_menu
from Imervue.menu.language_menu import build_language_menu
from Imervue.menu.recent_menu import rebuild_recent_menu
from Imervue.menu.sort_menu import build_sort_menu
from Imervue.menu.tip_menu import build_tip_menu
from Imervue.multi_language.language_wrapper import language_wrapper
from Imervue.user_settings.user_setting_dict import write_user_setting, read_user_setting, user_setting_dict


class ImervueMainWindow(QMainWindow):
    def __init__(self, debug: bool = False):
        super().__init__()

        self.setWindowTitle("Imervue")

        # Windows 平台設定 AppUserModelID
        # Set AppUserModelID for Windows platform
        self.id = "Imervue"
        try:
            from ctypes import windll
            windll.shell32.SetCurrentProcessExplicitAppUserModelID(self.id)
        except (ImportError, AttributeError):
            pass

        read_user_setting()
        last_folder = user_setting_dict.get("user_last_folder", "")

        if last_folder and Path(last_folder).is_dir():
            QTimer.singleShot(
                0,
                lambda: self._open_startup_folder(last_folder)
            )

        # 語言支援
        # Language support
        self.language_wrapper = language_wrapper
        self.language_wrapper.reset_language(user_setting_dict.get("language", "English"))

        self.icon_path = _app_icon_path()
        self.icon = QIcon(str(self.icon_path))
        self.setWindowIcon(self.icon)

        # ===== 中央 Splitter =====
        splitter = QSplitter()
        self.setCentralWidget(splitter)

        # ===== 左側檔案樹 =====
        self.model = QFileSystemModel()
        # 只篩選圖片格式 + 資料夾，隱藏不符合的檔案
        self.model.setNameFilters([
            "*.png", "*.jpg", "*.jpeg", "*.bmp", "*.tiff", "*.tif", "*.webp",
            "*.gif", "*.apng", "*.svg",
            "*.cr2", "*.nef", "*.arw", "*.dng", "*.raf", "*.orf",
        ])
        self.model.setNameFilterDisables(False)

        # 設定起始路徑：有上次資料夾就用，否則從根開始
        start_path = last_folder if last_folder and Path(last_folder).is_dir() else ""
        self.model.setRootPath(start_path)

        self.tree = QTreeView()
        self.tree.setModel(self.model)
        self.tree.setRootIndex(self.model.index(start_path))
        # 只顯示「名稱」欄，隱藏大小/類型/日期（省掉大量 stat() 呼叫）
        for col in (1, 2, 3):
            self.tree.hideColumn(col)
        self.tree.header().setStretchLastSection(True)
        self.tree.setColumnWidth(0, 400)
        self.tree.setUniformRowHeights(True)
        self.tree.setAnimated(False)
        self.tree.clicked.connect(self.on_file_clicked)

        # 目錄載入完成後確保樹狀圖更新
        self.model.directoryLoaded.connect(self._on_directory_loaded)

        # ===== 右側 GPU Viewer + Label =====
        right_widget = QWidget()
        right_layout = QVBoxLayout()
        right_layout.setContentsMargins(0, 0, 0, 0)
        right_layout.setSpacing(0)
        right_widget.setLayout(right_layout)

        self.filename_label = QLabel(language_wrapper.language_word_dict.get("main_window_current_filename"))
        self.filename_label.setMinimumHeight(16)  # 保證有高度
        self.filename_label.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        self.filename_label.setAlignment(Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter)
        self.filename_label.setWordWrap(False)
        self.filename_label.setTextInteractionFlags(Qt.TextInteractionFlag.TextSelectableByMouse)
        right_layout.addWidget(self.filename_label, stretch=0)

        # GPU Viewer + EXIF Sidebar 水平排列
        from Imervue.gpu_image_view.gpu_image_view import GPUImageView
        self.viewer = GPUImageView(main_window=self)

        self.exif_sidebar = ExifSidebar(self)

        viewer_row = QSplitter(Qt.Orientation.Horizontal)
        viewer_row.addWidget(self.viewer)
        viewer_row.addWidget(self.exif_sidebar)
        viewer_row.setStretchFactor(0, 1)
        viewer_row.setStretchFactor(1, 0)
        right_layout.addWidget(viewer_row, stretch=1)

        def _on_name_changed(name):
            self.filename_label.setText(
                language_wrapper.language_word_dict.get("main_window_current_filename_format").format(name=name))
            self.exif_sidebar.update_info()

        self.viewer.on_filename_changed = _on_name_changed

        splitter.addWidget(self.tree)
        splitter.addWidget(right_widget)
        splitter.setStretchFactor(1, 1)
        splitter.setSizes([400, 1000])

        # ===== 狀態列 =====
        self._status_bar = QStatusBar()
        self.setStatusBar(self._status_bar)
        self._status_label = QLabel("")
        self._progress_bar = QProgressBar()
        self._progress_bar.setFixedWidth(180)
        self._progress_bar.setVisible(False)
        self._status_bar.addWidget(self._status_label, stretch=1)
        self._status_bar.addPermanentWidget(self._progress_bar)

        # ===== Toast 通知 =====
        self.toast = ToastManager(self.viewer)

        # ===== 選單列 =====
        self.create_menu()

        _init_plugin_system_example(self)

        # ===== 資料夾監控 =====
        self._folder_watcher = QFileSystemWatcher(self)
        self._folder_watcher.directoryChanged.connect(self._on_watched_folder_changed)
        self._folder_refresh_timer = QTimer(self)
        self._folder_refresh_timer.setSingleShot(True)
        self._folder_refresh_timer.setInterval(500)  # 去抖動 500ms
        self._folder_refresh_timer.timeout.connect(self._do_folder_refresh)

        # ===== Debug =====
        # Debug 模式下自動關閉
        # Auto close in debug mode
        if debug:
            self.debug_timer = QTimer()
            self.debug_timer.setInterval(10000)
            self.debug_timer.timeout.connect(self.debug_close)
            self.debug_timer.start()

    # ==========================
    # 選單
    # ==========================
    def create_menu(self):
        build_file_menu(self)
        build_sort_menu(self)
        build_filter_menu(self)
        build_language_menu(self)
        # Plugin menu is built after plugin system init (in integration_guide.py)
        build_tip_menu(self)

    # ==========================
    # 狀態列
    # ==========================
    def set_status(self, text: str):
        self._status_label.setText(text)

    def show_progress(self, current: int, total: int):
        self._progress_bar.setVisible(True)
        self._progress_bar.setMaximum(total)
        self._progress_bar.setValue(current)
        self._status_label.setText(
            language_wrapper.language_word_dict.get(
                "status_loading_progress", "Loading {current}/{total}..."
            ).format(current=current, total=total)
        )
        if current >= total:
            QTimer.singleShot(800, self._hide_progress)

    def _hide_progress(self):
        self._progress_bar.setVisible(False)
        self._status_label.setText(
            language_wrapper.language_word_dict.get("status_ready", "Ready")
        )

    def change_tile_size(self, size):
        if size != "None":
            self.viewer.thumbnail_size = size
        else:
            self.viewer.thumbnail_size = None

        # 如果目前有載入圖片，重新刷新
        if self.viewer.model.images:
            if len(self.viewer.model.images) > 1:
                self.viewer.clear_tile_grid()
                self.viewer.load_tile_grid_async(image_paths=self.viewer.model.images)
            else:
                self.viewer.tile_grid_mode = False
                self.viewer.load_deep_zoom_image(self.viewer.model.images[0])

    # ==========================
    # 資料夾監控
    # ==========================
    def watch_folder(self, folder: str):
        """監控指定資料夾，發生變更時自動刷新"""
        dirs = self._folder_watcher.directories()
        if dirs:
            self._folder_watcher.removePaths(dirs)
        if folder:
            self._folder_watcher.addPath(folder)

    def _on_watched_folder_changed(self, _path: str):
        """資料夾內容變更 → 啟動去抖動計時器"""
        self._folder_refresh_timer.start()

    def _do_folder_refresh(self):
        """去抖動後重新掃描資料夾並更新 tile grid"""
        viewer = self.viewer
        if not viewer.model.images:
            return
        # 取得目前資料夾
        first = viewer.model.images[0]
        folder = str(Path(first).parent)
        if not Path(folder).is_dir():
            return

        from Imervue.gpu_image_view.images.image_loader import _scan_images
        new_images = _scan_images(folder)

        if set(new_images) != set(viewer.model.images):
            viewer.model.images = new_images
            if viewer.tile_grid_mode:
                viewer.clear_tile_grid()
                viewer.load_tile_grid_async(image_paths=new_images)
            viewer.update()

    # ==========================
    # 點擊檔案
    # ==========================
    def on_file_clicked(self, index):
        path = self.model.filePath(index)

        if Path(path).is_dir():
            # 點擊資料夾 → 導航進入並載入圖片
            self.model.setRootPath(path)
            self.tree.setRootIndex(self.model.index(path))
            self.viewer.clear_tile_grid()
            open_path(main_gui=self.viewer, path=path)
            self.filename_label.setText(
                language_wrapper.language_word_dict.get(
                    "main_window_current_folder_format"
                ).format(path=path)
            )
            self.watch_folder(path)
        elif Path(path).is_file():
            self.viewer.clear_tile_grid()
            open_path(main_gui=self.viewer, path=path)

        rebuild_recent_menu(self)

    def _on_directory_loaded(self, directory: str):
        """QFileSystemModel 目錄載入完成回調，確保樹狀圖正確顯示"""
        idx = self.model.index(directory)
        if idx.isValid():
            self.tree.update(idx)

    def _open_startup_folder(self, folder: str):
        self.model.setRootPath(folder)
        self.tree.setRootIndex(self.model.index(folder))
        open_path(main_gui=self.viewer, path=folder)
        self.filename_label.setText(
            language_wrapper.language_word_dict.get(
                "main_window_current_folder_format"
            ).format(path=folder)
        )
        self.watch_folder(folder)

    def closeEvent(self, event):
        # 最優先：儲存使用者設定（在任何可能失敗的操作之前）
        try:
            write_user_setting()
        except Exception:
            pass

        try:
            commit_pending_deletions(self.viewer)
        except Exception:
            pass

        # Plugin hook: app closing
        if hasattr(self, "plugin_manager"):
            try:
                self.plugin_manager.dispatch_app_closing(self)
                self.plugin_manager.unload_all()
            except Exception:
                pass

        event.accept()
        super().closeEvent(event)
        QApplication.instance().quit()

    @classmethod
    def debug_close(cls) -> None:
        """Debug 模式下強制退出 / Force exit in debug mode"""
        sys.exit(0)


if __name__ == "__main__":
    from Imervue.system.log_setup import setup_logging
    setup_logging()

    # HiDPI 支援
    QApplication.setHighDpiScaleFactorRoundingPolicy(
        Qt.HighDpiScaleFactorRoundingPolicy.PassThrough
    )
    app = QApplication(sys.argv)
    app.setQuitOnLastWindowClosed(False)
    window = ImervueMainWindow()
    window.showMaximized()
    sys.exit(app.exec())



class ImageModel:
    def __init__(self):
        self.images: list[str] = []
        self.deleted_history = []

    def set_images(self, paths: list[str]):
        self.images = paths.copy()

    def delete_images(self, paths: list[str]):
        items_to_delete = []
        for p in paths:
            if p in self.images:
                idx = self.images.index(p)
                items_to_delete.append((p, idx))
        # 從後往前刪除以保持 index 正確
        deleted = []
        for p, idx in sorted(items_to_delete, key=lambda x: x[1], reverse=True):
            self.images.pop(idx)
            deleted.append((p, idx))
        return deleted

    def restore_images(self, deleted_items):
        for path, index in sorted(deleted_items, key=lambda x: x[1]):
            self.images.insert(index, path)



from __future__ import annotations

from pathlib import Path

from PySide6.QtWidgets import QMenu

from Imervue.gpu_image_view.images.image_loader import open_path
from Imervue.multi_language.language_wrapper import language_wrapper
from Imervue.user_settings.recent_image import clear_recent, add_recent_folder, add_recent_image
from Imervue.user_settings.user_setting_dict import user_setting_dict


def build_recent_menu(ui_we_want_to_set, menu: QMenu):
    # ===== Recent =====
    recent_menu = menu.addMenu(language_wrapper.language_word_dict.get("recent_menu_title"))

    recent_folder_menu = recent_menu.addMenu("Recent Folders")
    recent_image_menu = recent_menu.addMenu("Recent Images")

    ui_we_want_to_set._recent_folder_menu = recent_folder_menu
    ui_we_want_to_set._recent_image_menu = recent_image_menu
    ui_we_want_to_set._recent_menu = recent_menu

    recent_menu.addSeparator()

    clear_action = recent_menu.addAction("Clear Recent")
    clear_action.triggered.connect(
        lambda: handle_clear_recent(ui_we_want_to_set)
    )

    rebuild_recent_menu(ui_we_want_to_set)


def rebuild_recent_menu(ui_we_want_to_set):
    folder_menu = ui_we_want_to_set._recent_folder_menu
    image_menu = ui_we_want_to_set._recent_image_menu

    folder_menu.clear()
    image_menu.clear()

    # ===== Folders =====
    valid_folders = []
    for path in user_setting_dict.get("user_recent_folders", []):
        if Path(path).is_dir():
            valid_folders.append(path)

            icon = ui_we_want_to_set.model.fileIcon(
                ui_we_want_to_set.model.index(path)
            )
            action = folder_menu.addAction(icon, path)
            action.triggered.connect(
                lambda checked, p=path: open_recent(ui_we_want_to_set, p)
            )

    user_setting_dict["user_recent_folders"] = valid_folders

    if not valid_folders:
        empty = folder_menu.addAction("(Empty)")
        empty.setEnabled(False)

    # ===== Images =====
    valid_images = []
    for path in user_setting_dict.get("user_recent_images", []):
        if Path(path).is_file():
            valid_images.append(path)

            action = image_menu.addAction(path)
            action.triggered.connect(
                lambda checked, p=path: open_recent(ui_we_want_to_set, p)
            )

    user_setting_dict["user_recent_images"] = valid_images

    if not valid_images:
        empty = image_menu.addAction("(Empty)")
        empty.setEnabled(False)


def handle_clear_recent(ui_we_want_to_set):
    clear_recent()
    rebuild_recent_menu(ui_we_want_to_set)


def open_recent(ui_we_want_to_set, path: str):

    if not Path(path).exists():
        return

    # 更新檔案樹定位
    if Path(path).is_dir():
        ui_we_want_to_set.model.setRootPath(path)
        ui_we_want_to_set.tree.setRootIndex(ui_we_want_to_set.model.index(path))
    else:
        parent = str(Path(path).parent)
        ui_we_want_to_set.model.setRootPath(parent)
        ui_we_want_to_set.tree.setRootIndex(ui_we_want_to_set.model.index(parent))

    ui_we_want_to_set.viewer.clear_tile_grid()
    open_path(main_gui=ui_we_want_to_set.viewer, path=path)

    if Path(path).is_dir():
        add_recent_folder(path)
        user_setting_dict["user_last_folder"] = path
    else:
        add_recent_image(path)
        user_setting_dict["user_last_folder"] = str(Path(path).parent)

    rebuild_recent_menu(ui_we_want_to_set)



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from PySide6.QtCore import QRunnable, Signal, QObject

from Imervue.image.pyramid import DeepZoomImage
from Imervue.image.tile_manager import TileManager

if TYPE_CHECKING:
    from Imervue.gpu_image_view.gpu_image_view import GPUImageView

import imageio
import numpy as np
import rawpy
from PIL import Image


def load_image_file(path, thumbnail=False):
    """
    支援一般圖片 + RAW 檔案
    thumbnail=True 時會優先使用 RAW embedded preview 或 half_size
    回傳 numpy RGBA
    """
    ext = Path(path).suffix.lower()

    raw_exts = {".cr2", ".nef", ".arw", ".dng", ".raf", ".orf"}

    # ===== RAW 檔案 =====
    if ext in raw_exts:
        with rawpy.imread(path) as raw:

            # ---- Thumbnail 模式 ----
            if thumbnail:
                # 優先讀 embedded preview
                try:
                    thumb = raw.extract_thumb()
                    if thumb.format == rawpy.ThumbFormat.JPEG:
                        img = imageio.v3.imread(thumb.data)
                    elif thumb.format == rawpy.ThumbFormat.BITMAP:
                        img = thumb.data
                    else:
                        raise Exception("No valid embedded preview")
                except Exception:
                    # fallback 用 half_size
                    img = raw.postprocess(
                        half_size=True,
                        use_camera_wb=True,
                        output_bps=8
                    )
            else:
                # ---- 全解析 ----
                img = raw.postprocess(
                    use_camera_wb=True,
                    no_auto_bright=False,
                    output_bps=8
                )

        img_data = img

    # ===== SVG 向量圖 =====
    elif ext == ".svg":
        img_data = _load_svg(path, thumbnail=thumbnail)

    # ===== 一般圖片 =====
    else:
        img = Image.open(path).convert("RGBA")
        img_data = np.array(img)

    # ===== 灰階 → RGB =====
    if img_data.ndim == 2:
        img_data = np.stack([img_data, img_data, img_data], axis=2)

    # ===== 補 alpha =====
    if img_data.shape[2] == 3:
        alpha = np.ones((*img_data.shape[:2], 1), dtype=np.uint8) * 255
        img_data = np.concatenate([img_data, alpha], axis=2)

    return img_data


# ================================================================
# DeepZoom 非同步載入 Worker
# ================================================================

class _DeepZoomWorkerSignals(QObject):
    finished = Signal(object, str)  # (DeepZoomImage, path)


class LoadDeepZoomWorker(QRunnable):
    """在背景執行緒載入全解析度圖片並建立金字塔，避免凍結 UI"""

    def __init__(self, path: str):
        super().__init__()
        self.path = path
        self.signals = _DeepZoomWorkerSignals()
        self._abort = False

    def abort(self):
        self._abort = True

    def run(self):
        if self._abort:
            return
        try:
            img_data = load_image_file(self.path, thumbnail=False)
            if self._abort:
                return
            dzi = DeepZoomImage(img_data)
            del img_data  # 釋放原始圖片記憶體，金字塔已持有降採樣副本
            if not self._abort:
                self.signals.finished.emit(dzi, self.path)
        except Exception as e:
            print(f"DeepZoom load failed: {self.path} - {e}")


# ================================================================
# 開啟路徑（資料夾或檔案）
# ================================================================

_SUPPORTED_EXTS = {
    ".png", ".jpg", ".jpeg", ".bmp", ".tiff", ".tif", ".webp",
    ".gif", ".apng", ".svg",
    ".cr2", ".nef", ".arw", ".dng", ".raf", ".orf",
}


def _load_svg(path: str, thumbnail: bool = False) -> np.ndarray:
    """Render SVG to RGBA numpy array using QSvgRenderer."""
    from PySide6.QtSvg import QSvgRenderer
    from PySide6.QtGui import QImage, QPainter
    from PySide6.QtCore import Qt, QSize

    renderer = QSvgRenderer(path)
    if not renderer.isValid():
        raise ValueError(f"Invalid SVG file: {path}")

    size = renderer.defaultSize()
    if size.isEmpty():
        size = QSize(1024, 1024)

    if thumbnail:
        # Scale down for thumbnail
        max_dim = 512
        if size.width() > max_dim or size.height() > max_dim:
            size.scale(max_dim, max_dim, Qt.AspectRatioMode.KeepAspectRatio)

    img = QImage(size, QImage.Format.Format_ARGB32_Premultiplied)
    img.fill(Qt.GlobalColor.transparent)
    painter = QPainter(img)
    renderer.render(painter)
    painter.end()

    # QImage → numpy (BGRA → RGBA)
    w, h = img.width(), img.height()
    ptr = img.bits()
    arr = np.frombuffer(ptr, dtype=np.uint8).reshape((h, w, 4)).copy()
    # BGRA → RGBA
    arr[:, :, [0, 2]] = arr[:, :, [2, 0]]
    return arr


def _scan_images(directory: str) -> list[str]:
    """快速掃描資料夾中的圖片，先篩選再排序，使用 os.scandir 提升效能"""
    import os
    result = []
    try:
        with os.scandir(directory) as it:
            for entry in it:
                if entry.is_file(follow_symlinks=False):
                    ext = os.path.splitext(entry.name)[1].lower()
                    if ext in _SUPPORTED_EXTS:
                        result.append(entry.path)
    except OSError:
        return []
    result.sort(key=lambda p: os.path.basename(p).lower())
    return result


def open_path(main_gui: GPUImageView, path: str):

    from Imervue.user_settings.user_setting_dict import user_setting_dict
    from Imervue.user_settings.recent_image import add_recent_folder, add_recent_image

    path_obj = Path(path)

    if path_obj.is_dir():

        images = _scan_images(str(path_obj))

        if not images:
            return

        main_gui.current_index = 0
        main_gui._unfiltered_images = list(images)
        main_gui.load_tile_grid_async(images)

        # 記錄最近開啟的資料夾
        add_recent_folder(str(path_obj))
        user_setting_dict["user_last_folder"] = str(path_obj)

        # Plugin hook: folder opened
        if hasattr(main_gui.main_window, "plugin_manager"):
            main_gui.main_window.plugin_manager.dispatch_folder_opened(str(path_obj), images, main_gui)

    elif path_obj.is_file() and path_obj.suffix.lower() in _SUPPORTED_EXTS:

        dir_path = path_obj.parent

        images = _scan_images(str(dir_path))

        if not images:
            return

        main_gui.model.set_images(images)
        try:
            main_gui.current_index = images.index(str(path_obj))
        except ValueError:
            # 路徑格式可能不同（正斜線/反斜線），用 normpath 比對
            import os
            norm = os.path.normpath(str(path_obj))
            for i, p in enumerate(images):
                if os.path.normpath(p) == norm:
                    main_gui.current_index = i
                    break
            else:
                main_gui.current_index = 0

        main_gui.tile_grid_mode = False
        main_gui.load_deep_zoom_image(str(path_obj))

        # 記錄最近開啟的檔案
        add_recent_image(str(path_obj))
        user_setting_dict["user_last_folder"] = str(dir_path)

        # Plugin hook: image loaded
        if hasattr(main_gui.main_window, "plugin_manager"):
            main_gui.main_window.plugin_manager.dispatch_image_loaded(str(path_obj), main_gui)



from __future__ import annotations

from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any

from PIL import Image
from PIL.ExifTags import TAGS
from PySide6.QtWidgets import QMessageBox

from Imervue.gpu_image_view.images.image_loader import load_image_file
from Imervue.multi_language.language_wrapper import language_wrapper

if TYPE_CHECKING:
    from Imervue.gpu_image_view.gpu_image_view import GPUImageView


# ==========================================================
# 取得滑鼠位置圖片資訊
# ==========================================================

def get_image_info_at_pos(main_gui: GPUImageView, position):
    mx, my = position.x(), position.y()

    # ===== Tile Grid 模式 =====
    if main_gui.tile_grid_mode:
        for x0, y0, x1, y1, path in main_gui.tile_rects:
            if x0 <= mx <= x1 and y0 <= my <= y1:
                return build_image_info(main_gui, Path(path))

    # ===== DeepZoom 模式 =====
    if main_gui.deep_zoom:
        if 0 <= main_gui.current_index < len(main_gui.model.images):
            path = Path(main_gui.model.images[main_gui.current_index])
            return build_image_info(main_gui, path)

    return None


# ==========================================================
# 建立圖片資訊
# ==========================================================

def build_image_info(main_gui: GPUImageView, path: Path) -> dict[str, Any]:
    info: dict[str, Any] = {}

    try:
        stat = path.stat()

        info["filename"] = path.name
        info["full_path"] = str(path)
        info["file_size_mb"] = round(stat.st_size / (1024 * 1024), 2)

        # ===== 檔案時間 =====
        ctime, mtime = get_file_times(path)
        info["created_time"] = ctime
        info["modified_time"] = mtime

        # ===== 讀取尺寸 =====
        cache_key = str(path)

        if cache_key in main_gui.tile_cache:
            img = main_gui.tile_cache[cache_key]
        else:
            img = load_image_file(path, thumbnail=True)

        h, w = img.shape[:2]
        info["width"] = w
        info["height"] = h

        # ===== EXIF =====
        exif = get_exif_data(path)
        info["exif_text"] = format_exif_info(exif)

    except Exception as e:
        info["error"] = str(e)

    return info


# ==========================================================
# 顯示 Dialog
# ==========================================================

def show_image_info_dialog(main_gui: GPUImageView, info: dict[str, Any]):
    if not info:
        return

    if "error" in info:
        QMessageBox.warning(main_gui, "Image Info Error", info["error"])
        return

    text = (
            language_wrapper.language_word_dict.get("image_info_filename").format(info=info["filename"]) +
            language_wrapper.language_word_dict.get("image_info_fullpath").format(full_path=info["full_path"]) +
            language_wrapper.language_word_dict.get(
                "image_info_image_size").format(width=info["width"], height=info["height"]) +
            language_wrapper.language_word_dict.get(
                "image_info_file_size").format(file_size_mb=info["file_size_mb"]) +
            language_wrapper.language_word_dict.get(
                "image_info_file_created_time").format(created_time=info["created_time"]) +
            language_wrapper.language_word_dict.get(
                "image_info_file_modified_time").format(modified_time=info["modified_time"]) +
            "\n====== EXIF ======\n"
            f"{info['exif_text']}"
    )

    QMessageBox.information(main_gui, language_wrapper.language_word_dict.get("image_info_messagebox_title"), text)


# ==========================================================
# 檔案時間
# ==========================================================

def get_file_times(path: Path):
    stat = path.stat()

    mtime = datetime.fromtimestamp(stat.st_mtime)

    try:
        ctime = datetime.fromtimestamp(stat.st_ctime)
    except Exception:
        ctime = None

    return ctime, mtime


# ==========================================================
# EXIF
# ==========================================================

def get_exif_data(path: Path):
    try:
        with Image.open(path) as img:
            exif_raw = img._getexif()

        if not exif_raw:
            return {}

        return {
            TAGS.get(tag, tag): value
            for tag, value in exif_raw.items()
        }

    except Exception:
        return {}


def format_exif_info(exif: dict):
    if not exif:
        return "No EXIF data"

    def get(key):
        return exif.get(key, "N/A")

    return (
        language_wrapper.language_word_dict.get(
            "image_info_exif_datatime_original").format(DateTimeOriginal=get("DateTimeOriginal")) +
        language_wrapper.language_word_dict.get(
            "image_info_exif_camera_model").format(Make=get("Make"), Model=get("Model")) +
        language_wrapper.language_word_dict.get(
            "image_info_exif_camera_lens_model").format(LensModel=get("LensModel")) +
        language_wrapper.language_word_dict.get(
            "image_info_exif_camera_focal_length").format(FocalLength=get("FocalLength")) +
        language_wrapper.language_word_dict.get(
            "image_info_exif_camera_fnumber").format(FNumber=get("FNumber")) +
        language_wrapper.language_word_dict.get(
            "image_info_exif_exposure_time").format(ExposureTime=get("ExposureTime")) +
        language_wrapper.language_word_dict.get(
            "image_info_exif_iso").format(ISOSpeedRatings=get("ISOSpeedRatings"))
    )



"""Bookmark / Collection management for cross-folder image collections."""

from __future__ import annotations

MAX_BOOKMARKS = 500


def get_bookmarks() -> list[str]:
    from Imervue.user_settings.user_setting_dict import user_setting_dict
    return user_setting_dict.get("bookmarks", [])


def add_bookmark(path: str) -> bool:
    from Imervue.user_settings.user_setting_dict import user_setting_dict
    bookmarks = user_setting_dict.setdefault("bookmarks", [])
    if path in bookmarks:
        return False
    bookmarks.append(path)
    if len(bookmarks) > MAX_BOOKMARKS:
        bookmarks[:] = bookmarks[-MAX_BOOKMARKS:]
    return True


def remove_bookmark(path: str) -> bool:
    from Imervue.user_settings.user_setting_dict import user_setting_dict
    bookmarks = user_setting_dict.get("bookmarks", [])
    if path in bookmarks:
        bookmarks.remove(path)
        return True
    return False


def is_bookmarked(path: str) -> bool:
    from Imervue.user_settings.user_setting_dict import user_setting_dict
    return path in user_setting_dict.get("bookmarks", [])


def clear_bookmarks():
    from Imervue.user_settings.user_setting_dict import user_setting_dict
    user_setting_dict["bookmarks"] = []



"""
Toast 通知系統
Lightweight toast/snackbar notifications displayed at the bottom of the viewer.
"""
from __future__ import annotations

from PySide6.QtCore import Qt, QTimer, QPropertyAnimation, QEasingCurve
from PySide6.QtWidgets import QLabel, QWidget, QGraphicsOpacityEffect


class ToastWidget(QLabel):
    """半透明浮動通知標籤"""

    _STYLE = {
        "info": "background: rgba(50,50,50,200); color: #eee;",
        "success": "background: rgba(30,100,50,210); color: #eee;",
        "warning": "background: rgba(160,120,10,210); color: #fff;",
        "error": "background: rgba(160,30,30,210); color: #fff;",
    }

    def __init__(self, parent: QWidget):
        super().__init__(parent)
        self.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.setWordWrap(True)
        self.setVisible(False)
        self.setAttribute(Qt.WidgetAttribute.WA_TransparentForMouseEvents)

        self._opacity = QGraphicsOpacityEffect(self)
        self._opacity.setOpacity(0.0)
        self.setGraphicsEffect(self._opacity)

        self._hide_timer = QTimer(self)
        self._hide_timer.setSingleShot(True)
        self._hide_timer.timeout.connect(self._fade_out)

        self._anim: QPropertyAnimation | None = None

    def show_message(self, text: str, level: str = "info", duration_ms: int = 2500):
        base = self._STYLE.get(level, self._STYLE["info"])
        self.setStyleSheet(
            f"{base} border-radius: 6px; padding: 8px 18px; font-size: 13px;"
        )
        self.setText(text)
        self.adjustSize()
        self._reposition()
        self.setVisible(True)

        # 淡入
        if self._anim:
            self._anim.stop()
        self._anim = QPropertyAnimation(self._opacity, b"opacity")
        self._anim.setDuration(200)
        self._anim.setStartValue(0.0)
        self._anim.setEndValue(1.0)
        self._anim.start()

        self._hide_timer.start(duration_ms)

    def _fade_out(self):
        if self._anim:
            self._anim.stop()
        self._anim = QPropertyAnimation(self._opacity, b"opacity")
        self._anim.setDuration(400)
        self._anim.setStartValue(self._opacity.opacity())
        self._anim.setEndValue(0.0)
        self._anim.setEasingCurve(QEasingCurve.Type.InQuad)
        self._anim.finished.connect(lambda: self.setVisible(False))
        self._anim.start()

    def _reposition(self):
        parent = self.parentWidget()
        if parent:
            w = min(self.sizeHint().width() + 36, parent.width() - 40)
            self.setFixedWidth(w)
            x = (parent.width() - w) // 2
            y = parent.height() - self.sizeHint().height() - 40
            self.move(x, max(y, 10))


class ToastManager:
    """全局 toast 管理器（每個主視窗一個）"""

    def __init__(self, parent: QWidget):
        self._toast = ToastWidget(parent)

    def info(self, text: str, duration_ms: int = 2500):
        self._toast.show_message(text, "info", duration_ms)

    def success(self, text: str, duration_ms: int = 2500):
        self._toast.show_message(text, "success", duration_ms)

    def warning(self, text: str, duration_ms: int = 3000):
        self._toast.show_message(text, "warning", duration_ms)

    def error(self, text: str, duration_ms: int = 4000):
        self._toast.show_message(text, "error", duration_ms)



from __future__ import annotations

import logging
from pathlib import Path

from PIL import Image

logger = logging.getLogger("Imervue.lossless_rotate")

# EXIF Orientation tag value after a 90-degree clockwise rotation.
# Maps current_orientation -> new_orientation.
CW_ORIENTATION_MAP: dict[int, int] = {
    1: 6,
    6: 3,
    3: 8,
    8: 1,
    2: 7,
    7: 4,
    4: 5,
    5: 2,
}

# Counter-clockwise is the inverse of clockwise.
CCW_ORIENTATION_MAP: dict[int, int] = {v: k for k, v in CW_ORIENTATION_MAP.items()}

# JPEG file signatures (SOI marker)
_JPEG_EXTENSIONS: set[str] = {".jpg", ".jpeg", ".jpe", ".jfif"}


def _is_jpeg(file_path: str) -> bool:
    return Path(file_path).suffix.lower() in _JPEG_EXTENSIONS


def _rotate_via_exif(file_path: str, clockwise: bool) -> bool:
    """Rotate a JPEG by modifying its EXIF Orientation tag (truly lossless).

    Returns True on success, False if piexif is unavailable or an error occurs.
    """
    try:
        import piexif
    except ImportError:
        logger.debug("piexif not available; falling back to PIL rotation")
        return False

    try:
        exif_dict = piexif.load(file_path)

        # Read current orientation (default to 1 = normal if absent)
        current_orientation = exif_dict.get("0th", {}).get(
            piexif.ImageIFD.Orientation, 1
        )

        rotation_map = CW_ORIENTATION_MAP if clockwise else CCW_ORIENTATION_MAP
        new_orientation = rotation_map.get(current_orientation, 6 if clockwise else 8)

        exif_dict.setdefault("0th", {})[piexif.ImageIFD.Orientation] = new_orientation

        exif_bytes = piexif.dump(exif_dict)
        piexif.insert(exif_bytes, file_path)

        direction = "CW" if clockwise else "CCW"
        logger.info(
            f"Lossless EXIF rotate {direction}: orientation {current_orientation} -> "
            f"{new_orientation} for {file_path}"
        )
        return True
    except Exception as exc:
        logger.error(f"EXIF rotation failed for {file_path}: {exc}")
        return False


def _rotate_via_pil(file_path: str, clockwise: bool) -> bool:
    """Rotate an image using PIL transpose and re-save (lossy for compressed formats)."""
    try:
        img = Image.open(file_path)

        if clockwise:
            rotated = img.transpose(Image.Transpose.ROTATE_270)
        else:
            rotated = img.transpose(Image.Transpose.ROTATE_90)

        # Preserve original format
        fmt = img.format or Path(file_path).suffix.lstrip(".").upper()
        if fmt == "JPG":
            fmt = "JPEG"

        save_kwargs: dict = {}
        if fmt == "PNG":
            save_kwargs["compress_level"] = 6
        elif fmt == "JPEG":
            save_kwargs["quality"] = 95
            # Ensure RGB mode for JPEG
            if rotated.mode in ("RGBA", "P"):
                rotated = rotated.convert("RGB")
        elif fmt == "WEBP":
            save_kwargs["quality"] = 90

        rotated.save(file_path, format=fmt, **save_kwargs)

        direction = "CW" if clockwise else "CCW"
        logger.info(f"PIL rotate {direction}: {file_path}")
        return True
    except Exception as exc:
        logger.error(f"PIL rotation failed for {file_path}: {exc}")
        return False


def lossless_rotate(file_path: str, clockwise: bool = True) -> bool:
    """Rotate an image file by 90 degrees.

    For JPEG files, attempts a truly lossless rotation by modifying the EXIF
    Orientation tag via *piexif*.  Falls back to PIL transpose + re-save when
    piexif is not installed or for non-JPEG formats.

    Args:
        file_path: Absolute path to the image file.
        clockwise: If True rotate 90 degrees clockwise; otherwise counter-clockwise.

    Returns:
        True if the rotation was applied successfully, False otherwise.
    """
    if not Path(file_path).is_file():
        logger.error(f"File not found: {file_path}")
        return False

    # Attempt lossless EXIF rotation for JPEG files
    if _is_jpeg(file_path):
        if _rotate_via_exif(file_path, clockwise):
            return True
        # piexif unavailable or failed; fall through to PIL

    return _rotate_via_pil(file_path, clockwise)



"""
Centralized logging configuration for Imervue.

Call ``setup_logging()`` once at startup (before any other import that uses
``logging.getLogger``).  In frozen (PyInstaller) builds the log file is
written next to the .exe; in development it goes to the project root.
"""
from __future__ import annotations

import logging
import sys
from pathlib import Path


def setup_logging() -> None:
    """Configure the root ``Imervue`` logger to write to a file + stderr."""
    from Imervue.system.app_paths import app_dir

    log_path = app_dir() / "imervue.log"

    # Rotate: keep only the last run (overwrite)
    handler = logging.FileHandler(str(log_path), mode="w", encoding="utf-8")
    handler.setFormatter(logging.Formatter(
        "%(asctime)s [%(levelname)s] %(name)s: %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    ))

    root_logger = logging.getLogger("Imervue")
    root_logger.setLevel(logging.DEBUG)
    root_logger.addHandler(handler)

    # Also log to stderr when a console is available
    if not getattr(sys, "frozen", False):
        console = logging.StreamHandler()
        console.setFormatter(logging.Formatter(
            "[%(levelname)s] %(name)s: %(message)s"
        ))
        root_logger.addHandler(console)

    root_logger.info("Logging initialized — log file: %s", log_path)



from collections import OrderedDict

import numpy as np
from OpenGL.GL import (
    glDeleteTextures, glGenTextures, glTexImage2D, glBindTexture,
    glTexParameteri, glPixelStorei,
    GL_TEXTURE_2D, GL_RGBA, GL_UNSIGNED_BYTE,
    GL_TEXTURE_MIN_FILTER, GL_TEXTURE_MAG_FILTER,
    GL_UNPACK_ALIGNMENT, GL_LINEAR,
    GL_TEXTURE_WRAP_S, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE,
)


_MAX_TILE_CACHE = 256


class TileManager:
    def __init__(self, deep_zoom):
        self.deep_zoom = deep_zoom
        self.cache = OrderedDict()
        self._current_level = -1

    def clear(self):
        """釋放所有 GPU texture"""
        if self.cache:
            glDeleteTextures(list(self.cache.values()))
            self.cache.clear()

    def get_tile(self, level, tx, ty, tile_size):
        self._current_level = level

        key = (level, tx, ty)

        if key in self.cache:
            self.cache.move_to_end(key)
            return self.cache[key]

        # LRU 淘汰：優先淘汰不同 level 的 tile，再淘汰同 level 最舊的
        while len(self.cache) >= _MAX_TILE_CACHE:
            # 嘗試找不同 level 的最舊 tile
            evicted = False
            for k in list(self.cache):
                if k[0] != level:
                    glDeleteTextures([self.cache.pop(k)])
                    evicted = True
                    break
            if not evicted:
                _, tex = self.cache.popitem(last=False)
                glDeleteTextures([tex])

        image = self.deep_zoom.levels[level]
        h, w = image.shape[:2]

        x0 = tx * tile_size
        y0 = ty * tile_size
        x1 = min(x0 + tile_size, w)
        y1 = min(y0 + tile_size, h)

        if x0 >= w or y0 >= h:
            return None

        tile = image[y0:y1, x0:x1]

        tile = tile.astype(np.uint8)
        if tile.ndim == 2:
            tile = np.stack([tile, tile, tile], axis=2)
        if tile.shape[2] == 3:
            alpha = np.full((tile.shape[0], tile.shape[1], 1), 255, dtype=np.uint8)
            tile = np.concatenate([tile, alpha], axis=2)

        tex = glGenTextures(1)
        glBindTexture(GL_TEXTURE_2D, tex)
        glPixelStorei(GL_UNPACK_ALIGNMENT, 1)

        glTexImage2D(
            GL_TEXTURE_2D,
            0,
            GL_RGBA,
            tile.shape[1],
            tile.shape[0],
            0,
            GL_RGBA,
            GL_UNSIGNED_BYTE,
            tile
        )
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE)

        self.cache[key] = tex
        return tex



from __future__ import annotations

import os
import subprocess
import sys
from pathlib import Path
from typing import TYPE_CHECKING

from PySide6.QtGui import QClipboard
from PySide6.QtWidgets import QMenu, QApplication

from Imervue.gpu_image_view.actions.batch_ops import (
    open_batch_rename, open_batch_move, batch_rotate,
)
from Imervue.gpu_image_view.actions.compare_dialog import open_compare_dialog
from Imervue.gui.image_editor import open_image_editor
from Imervue.gpu_image_view.actions.lossless_rotate import lossless_rotate
from Imervue.gpu_image_view.actions.slideshow import open_slideshow_dialog
from Imervue.gui.export_dialog import open_export_dialog
from Imervue.gui.batch_export_dialog import open_batch_export
from Imervue.gui.gif_video_dialog import open_gif_video_dialog
from Imervue.gui.tag_album_dialog import (
    build_tag_submenu, build_album_submenu, build_batch_tag_album_submenu,
)
from Imervue.gpu_image_view.actions.keyboard_actions import (
    trash_current_image, trash_selected_tiles,
    rotate_current_image, copy_image_to_clipboard,
)
from Imervue.gpu_image_view.actions.select import switch_to_previous_image, switch_to_next_image
from Imervue.image.info import get_image_info_at_pos, show_image_info_dialog
from Imervue.multi_language.language_wrapper import language_wrapper
from Imervue.menu.recent_menu import build_recent_menu

if TYPE_CHECKING:
    from Imervue.gpu_image_view.gpu_image_view import GPUImageView


def right_click_context_menu(main_gui: GPUImageView, global_pos, local_pos):
    build_right_click_menu = QMenu(main_gui)

    go_to_parent_folder_action(main_gui=main_gui, menu=build_right_click_menu)
    switch_actions(main_gui=main_gui, menu=build_right_click_menu)

    build_right_click_menu.addSeparator()

    # === 新增快速操作 ===
    _show_in_explorer_action(main_gui, build_right_click_menu)
    _copy_path_action(main_gui, build_right_click_menu)
    _copy_image_action(main_gui, build_right_click_menu)

    build_right_click_menu.addSeparator()

    _rotate_actions(main_gui, build_right_click_menu)
    _edit_action(main_gui, build_right_click_menu)
    _batch_actions(main_gui, build_right_click_menu)
    _delete_action(main_gui, build_right_click_menu)
    _set_wallpaper_action(main_gui, build_right_click_menu)

    build_right_click_menu.addSeparator()

    _compare_action(main_gui, build_right_click_menu)
    _slideshow_action(main_gui, build_right_click_menu)

    build_right_click_menu.addSeparator()

    _export_action(main_gui, build_right_click_menu)
    _lossless_rotate_actions(main_gui, build_right_click_menu)

    build_right_click_menu.addSeparator()

    _bookmark_action(main_gui, build_right_click_menu)
    _tag_album_actions(main_gui, build_right_click_menu)

    build_right_click_menu.addSeparator()

    image_info_action(main_gui=main_gui, local_pos=local_pos, menu=build_right_click_menu)
    build_recent_menu(main_gui.main_window, build_right_click_menu)

    # Plugin hook: context menu
    if hasattr(main_gui.main_window, "plugin_manager"):
        main_gui.main_window.plugin_manager.dispatch_build_context_menu(build_right_click_menu, main_gui)

    if build_right_click_menu.actions():
        build_right_click_menu.exec(global_pos)


# ===========================
# 取得目前圖片路徑
# ===========================

def _current_image_path(main_gui: GPUImageView) -> str | None:
    images = main_gui.model.images
    if not images:
        return None
    if main_gui.deep_zoom and 0 <= main_gui.current_index < len(images):
        return images[main_gui.current_index]
    return None


# ===========================
# 在檔案總管顯示
# ===========================

def _show_in_explorer_action(main_gui: GPUImageView, menu: QMenu):
    path = _current_image_path(main_gui)
    if not path:
        return

    lang = language_wrapper.language_word_dict
    action = menu.addAction(lang.get("right_click_show_in_explorer", "Show in Explorer"))
    action.triggered.connect(lambda: _open_in_explorer(path))


def _open_in_explorer(path: str):
    try:
        if sys.platform == "win32":
            subprocess.Popen(["explorer", "/select,", os.path.normpath(path)])
        elif sys.platform == "darwin":
            subprocess.Popen(["open", "-R", path])
        else:
            subprocess.Popen(["xdg-open", str(Path(path).parent)])
    except Exception:
        pass


# ===========================
# 複製路徑
# ===========================

def _copy_path_action(main_gui: GPUImageView, menu: QMenu):
    path = _current_image_path(main_gui)
    if not path:
        return

    lang = language_wrapper.language_word_dict
    action = menu.addAction(lang.get("right_click_copy_path", "Copy Path"))
    def _do_copy_path():
        QApplication.clipboard().setText(path)
        if hasattr(main_gui.main_window, "toast"):
            main_gui.main_window.toast.info(
                language_wrapper.language_word_dict.get("toast_path_copied", "Path copied")
            )
    action.triggered.connect(_do_copy_path)


# ===========================
# 複製圖片
# ===========================

def _copy_image_action(main_gui: GPUImageView, menu: QMenu):
    if not main_gui.deep_zoom:
        return
    lang = language_wrapper.language_word_dict
    action = menu.addAction(lang.get("right_click_copy_image", "Copy Image"))
    action.triggered.connect(lambda: copy_image_to_clipboard(main_gui))


# ===========================
# 旋轉
# ===========================

def _rotate_actions(main_gui: GPUImageView, menu: QMenu):
    if not main_gui.deep_zoom:
        return

    lang = language_wrapper.language_word_dict

    cw_action = menu.addAction(lang.get("right_click_rotate_cw", "Rotate Clockwise"))
    cw_action.triggered.connect(lambda: rotate_current_image(main_gui, clockwise=True))

    ccw_action = menu.addAction(lang.get("right_click_rotate_ccw", "Rotate Counter-clockwise"))
    ccw_action.triggered.connect(lambda: rotate_current_image(main_gui, clockwise=False))


# ===========================
# 刪除（移至垃圾桶）
# ===========================

def _batch_actions(main_gui: GPUImageView, menu: QMenu):
    if not (main_gui.tile_grid_mode and main_gui.tile_selection_mode and main_gui.selected_tiles):
        return
    lang = language_wrapper.language_word_dict
    batch_menu = menu.addMenu(lang.get("right_click_batch", "Batch Operations"))

    rename_action = batch_menu.addAction(lang.get("batch_rename_title", "Batch Rename"))
    rename_action.triggered.connect(lambda: open_batch_rename(main_gui))

    move_action = batch_menu.addAction(lang.get("batch_move_title", "Move / Copy"))
    move_action.triggered.connect(lambda: open_batch_move(main_gui))

    rot_cw = batch_menu.addAction(lang.get("batch_rotate_cw", "Rotate All CW"))
    rot_cw.triggered.connect(lambda: batch_rotate(main_gui, list(main_gui.selected_tiles), 90))

    rot_ccw = batch_menu.addAction(lang.get("batch_rotate_ccw", "Rotate All CCW"))
    rot_ccw.triggered.connect(lambda: batch_rotate(main_gui, list(main_gui.selected_tiles), -90))

    batch_menu.addSeparator()

    export_action = batch_menu.addAction(lang.get("batch_export_title", "Batch Export"))
    export_action.triggered.connect(lambda: open_batch_export(main_gui))

    gif_action = batch_menu.addAction(lang.get("gif_video_title", "Create GIF / Video"))
    gif_action.triggered.connect(lambda: open_gif_video_dialog(main_gui))

    build_batch_tag_album_submenu(main_gui, list(main_gui.selected_tiles), batch_menu)


def _edit_action(main_gui: GPUImageView, menu: QMenu):
    path = _current_image_path(main_gui)
    if not path:
        return
    lang = language_wrapper.language_word_dict
    action = menu.addAction(lang.get("right_click_edit_image", "Edit Image"))
    action.triggered.connect(lambda: open_image_editor(main_gui))


def _delete_action(main_gui: GPUImageView, menu: QMenu):
    lang = language_wrapper.language_word_dict

    if main_gui.tile_grid_mode and main_gui.tile_selection_mode:
        action = menu.addAction(lang.get("right_click_menu_delete_selected"))
        action.triggered.connect(lambda: trash_selected_tiles(main_gui))

    if main_gui.deep_zoom:
        action = menu.addAction(lang.get("right_click_menu_delete_current"))
        action.triggered.connect(lambda: trash_current_image(main_gui))


# ===========================
# 設為桌布
# ===========================

def _set_wallpaper_action(main_gui: GPUImageView, menu: QMenu):
    path = _current_image_path(main_gui)
    if not path:
        return

    lang = language_wrapper.language_word_dict
    action = menu.addAction(lang.get("right_click_set_wallpaper", "Set as Wallpaper"))
    action.triggered.connect(lambda: _set_wallpaper(path))


def _set_wallpaper(path: str):
    try:
        if sys.platform == "win32":
            import ctypes
            SPI_SETDESKWALLPAPER = 0x0014
            SPIF_UPDATEINIFILE = 0x01
            SPIF_SENDCHANGE = 0x02
            ctypes.windll.user32.SystemParametersInfoW(
                SPI_SETDESKWALLPAPER, 0, os.path.normpath(path),
                SPIF_UPDATEINIFILE | SPIF_SENDCHANGE
            )
        elif sys.platform == "darwin":
            script = f'''
            tell application "Finder"
                set desktop picture to POSIX file "{path}"
            end tell
            '''
            subprocess.Popen(["osascript", "-e", script])
        else:
            # GNOME
            subprocess.Popen([
                "gsettings", "set", "org.gnome.desktop.background",
                "picture-uri", f"file://{path}"
            ])
    except Exception:
        pass


# ===========================
# 原有功能
# ===========================

def _slideshow_action(main_gui: GPUImageView, menu: QMenu):
    if not main_gui.model.images or len(main_gui.model.images) < 2:
        return
    lang = language_wrapper.language_word_dict

    if hasattr(main_gui, '_slideshow') and main_gui._slideshow and main_gui._slideshow.running:
        action = menu.addAction(lang.get("right_click_slideshow_stop", "Stop Slideshow"))
        action.triggered.connect(lambda: main_gui._slideshow.stop())
    else:
        action = menu.addAction(lang.get("right_click_slideshow", "Slideshow"))
        action.triggered.connect(lambda: open_slideshow_dialog(main_gui))


def _compare_action(main_gui: GPUImageView, menu: QMenu):
    if not main_gui.model.images or len(main_gui.model.images) < 2:
        return
    lang = language_wrapper.language_word_dict
    action = menu.addAction(lang.get("right_click_compare", "Compare Images"))
    action.triggered.connect(lambda: open_compare_dialog(main_gui))


def switch_actions(main_gui: GPUImageView, menu: QMenu):
    if main_gui.deep_zoom:
        next_image_action = menu.addAction(
            language_wrapper.language_word_dict.get("right_click_menu_next_image"))
        next_image_action.triggered.connect(
            lambda: switch_to_next_image(main_gui=main_gui)
        )
        previous_image_action = menu.addAction(
            language_wrapper.language_word_dict.get("right_click_menu_previous_image")
        )
        previous_image_action.triggered.connect(
            lambda: switch_to_previous_image(main_gui=main_gui)
        )


def go_to_parent_folder_action(main_gui: GPUImageView, menu: QMenu):
    images = main_gui.model.images
    if images:
        current_path = None
        if main_gui.deep_zoom and 0 <= main_gui.current_index < len(images):
            current_path = images[main_gui.current_index]
        elif main_gui.tile_grid_mode:
            current_path = images[0]

        if current_path:
            parent_folder = Path(current_path).parent.parent

            if parent_folder.exists():
                action = menu.addAction(
                    language_wrapper.language_word_dict.get("right_click_menu_go_to_parent_folder"))
                action.triggered.connect(
                    lambda: jump_to_folder(main_gui=main_gui, folder_path=parent_folder)
                )


def jump_to_folder(main_gui: GPUImageView, folder_path: Path):
    model = main_gui.main_window.model
    tree = main_gui.main_window.tree

    index = model.index(str(folder_path))
    if index.isValid():
        tree.setCurrentIndex(index)
        tree.scrollTo(index)
        tree.setRootIndex(index)


def image_info_action(main_gui: GPUImageView, local_pos, menu: QMenu):
    info = get_image_info_at_pos(main_gui=main_gui, position=local_pos)
    if info:
        action = menu.addAction(
            language_wrapper.language_word_dict.get("right_click_menu_image_info")
        )
        action.triggered.connect(
            lambda: image_info(main_gui=main_gui, info=info)
        )


def image_info(main_gui: GPUImageView, info):
    show_image_info_dialog(main_gui=main_gui, info=info)


# ===========================
# 匯出 / 另存為
# ===========================

def _export_action(main_gui: GPUImageView, menu: QMenu):
    path = _current_image_path(main_gui)
    if not path:
        return
    lang = language_wrapper.language_word_dict
    action = menu.addAction(lang.get("right_click_export", "Export / Save As..."))
    action.triggered.connect(lambda: open_export_dialog(main_gui))


# ===========================
# 無損旋轉
# ===========================

def _lossless_rotate_actions(main_gui: GPUImageView, menu: QMenu):
    path = _current_image_path(main_gui)
    if not path:
        return
    lang = language_wrapper.language_word_dict
    sub = menu.addMenu(lang.get("right_click_lossless_rotate", "Lossless Rotate"))

    cw = sub.addAction(lang.get("lossless_rotate_cw", "Lossless Rotate CW"))
    cw.triggered.connect(lambda: _do_lossless_rotate(main_gui, path, clockwise=True))

    ccw = sub.addAction(lang.get("lossless_rotate_ccw", "Lossless Rotate CCW"))
    ccw.triggered.connect(lambda: _do_lossless_rotate(main_gui, path, clockwise=False))


def _bookmark_action(main_gui: GPUImageView, menu: QMenu):
    path = _current_image_path(main_gui)
    if not path:
        return
    lang = language_wrapper.language_word_dict
    from Imervue.user_settings.bookmark import is_bookmarked
    if is_bookmarked(path):
        action = menu.addAction(lang.get("bookmark_remove_current", "Remove Bookmark"))
    else:
        action = menu.addAction(lang.get("bookmark_add_current", "Add Bookmark"))
    action.triggered.connect(lambda: main_gui._toggle_bookmark())


def _tag_album_actions(main_gui: GPUImageView, menu: QMenu):
    path = _current_image_path(main_gui)
    if not path:
        return
    build_tag_submenu(main_gui, path, menu)
    build_album_submenu(main_gui, path, menu)


def _do_lossless_rotate(main_gui: GPUImageView, path: str, clockwise: bool):
    lang = language_wrapper.language_word_dict
    ok = lossless_rotate(path, clockwise=clockwise)
    if ok:
        if hasattr(main_gui.main_window, "toast"):
            main_gui.main_window.toast.info(
                lang.get("lossless_rotate_done", "Lossless rotation saved!")
            )
        # 重新載入圖片以反映旋轉
        from Imervue.gpu_image_view.images.image_loader import open_path
        main_gui._clear_deep_zoom()
        open_path(main_gui=main_gui, path=path)
    else:
        if hasattr(main_gui.main_window, "toast"):
            main_gui.main_window.toast.info(
                lang.get("lossless_rotate_fail", "Lossless rotation failed")
            )



"""
GIF/影片製作
Create GIF or MP4 video from selected images.
"""
from __future__ import annotations

import logging
from pathlib import Path
from typing import TYPE_CHECKING

from PySide6.QtCore import Qt, QThread, Signal
from PySide6.QtWidgets import (
    QDialog, QVBoxLayout, QHBoxLayout, QLabel, QComboBox,
    QSpinBox, QPushButton, QFileDialog, QLineEdit, QProgressBar,
    QListWidget, QListWidgetItem, QGroupBox, QCheckBox,
)
from PIL import Image

from Imervue.multi_language.language_wrapper import language_wrapper

if TYPE_CHECKING:
    from Imervue.gpu_image_view.gpu_image_view import GPUImageView

logger = logging.getLogger("Imervue.gif_video")


class _CreateWorker(QThread):
    progress = Signal(int, int)
    finished = Signal(bool, str)  # success, message

    def __init__(self, paths, output_path, fmt, fps, width, height, loop):
        super().__init__()
        self._paths = paths
        self._output = output_path
        self._fmt = fmt
        self._fps = fps
        self._width = width
        self._height = height
        self._loop = loop

    def run(self):
        try:
            if self._fmt == "GIF":
                self._create_gif()
            else:
                self._create_video()
            self.finished.emit(True, self._output)
        except Exception as exc:
            logger.error(f"Create {self._fmt} failed: {exc}")
            self.finished.emit(False, str(exc))

    def _load_and_resize(self, path: str) -> Image.Image:
        if Path(path).suffix.lower() == ".svg":
            from Imervue.gpu_image_view.images.image_loader import _load_svg
            arr = _load_svg(path, thumbnail=False)
            img = Image.fromarray(arr)
        else:
            img = Image.open(path)

        if img.mode not in ("RGB", "RGBA"):
            img = img.convert("RGB")

        if self._width > 0 and self._height > 0:
            img = img.resize((self._width, self._height), Image.Resampling.LANCZOS)
        elif self._width > 0 or self._height > 0:
            img.thumbnail(
                (self._width or img.width, self._height or img.height),
                Image.Resampling.LANCZOS,
            )
        return img

    def _create_gif(self):
        frames = []
        total = len(self._paths)
        for i, path in enumerate(self._paths):
            img = self._load_and_resize(path)
            if img.mode == "RGBA":
                img = img.convert("RGB")
            frames.append(img)
            self.progress.emit(i + 1, total)

        if not frames:
            return

        duration = int(1000 / max(self._fps, 1))
        loop_val = 0 if self._loop else 1

        frames[0].save(
            self._output,
            format="GIF",
            save_all=True,
            append_images=frames[1:],
            duration=duration,
            loop=loop_val,
            optimize=True,
        )

    def _create_video(self):
        import subprocess
        import sys
        import tempfile
        import shutil

        # Check ffmpeg
        ffmpeg = shutil.which("ffmpeg")
        if not ffmpeg:
            self.finished.emit(False, "ffmpeg not found in PATH")
            return

        total = len(self._paths)
        tmpdir = tempfile.mkdtemp(prefix="imervue_video_")

        try:
            # Write frames as numbered images
            for i, path in enumerate(self._paths):
                img = self._load_and_resize(path)
                if img.mode == "RGBA":
                    img = img.convert("RGB")
                frame_path = Path(tmpdir) / f"frame_{i:06d}.png"
                img.save(str(frame_path), format="PNG")
                self.progress.emit(i + 1, total)

            # Use ffmpeg to combine
            pattern = str(Path(tmpdir) / "frame_%06d.png")
            cmd = [
                ffmpeg, "-y",
                "-framerate", str(self._fps),
                "-i", pattern,
                "-c:v", "libx264",
                "-pix_fmt", "yuv420p",
                "-preset", "fast",
                self._output,
            ]
            kw = {
                "capture_output": True,
                "encoding": "utf-8",
                "errors": "replace",
                "timeout": 300,
                "stdin": subprocess.DEVNULL,
            }
            if sys.platform == "win32":
                kw["creationflags"] = subprocess.CREATE_NO_WINDOW
            result = subprocess.run(cmd, **kw)
            if result.returncode != 0:
                self.finished.emit(False, f"ffmpeg error: {result.stderr[:200]}")
                return
        finally:
            shutil.rmtree(tmpdir, ignore_errors=True)


class GifVideoDialog(QDialog):
    def __init__(self, main_gui: GPUImageView, paths: list[str]):
        super().__init__(main_gui.main_window)
        self._gui = main_gui
        self._paths = list(paths)
        self._lang = language_wrapper.language_word_dict
        self._worker = None

        self.setWindowTitle(self._lang.get("gif_video_title", "Create GIF / Video"))
        self.setMinimumSize(520, 520)

        self._build_ui()

    def _build_ui(self):
        layout = QVBoxLayout(self)

        # Image list (reorderable)
        layout.addWidget(QLabel(
            self._lang.get("gif_video_order", "Drag to reorder (top = first frame):")
        ))
        self._list = QListWidget()
        self._list.setDragDropMode(QListWidget.DragDropMode.InternalMove)
        for p in self._paths:
            item = QListWidgetItem(Path(p).name)
            item.setData(Qt.ItemDataRole.UserRole, p)
            self._list.addItem(item)
        layout.addWidget(self._list)

        # Move up/down buttons
        order_row = QHBoxLayout()
        up_btn = QPushButton(self._lang.get("gif_video_move_up", "Move Up"))
        up_btn.clicked.connect(self._move_up)
        down_btn = QPushButton(self._lang.get("gif_video_move_down", "Move Down"))
        down_btn.clicked.connect(self._move_down)
        order_row.addWidget(up_btn)
        order_row.addWidget(down_btn)
        order_row.addStretch()
        layout.addLayout(order_row)

        # Settings
        settings = QGroupBox(self._lang.get("gif_video_settings", "Settings"))
        slay = QVBoxLayout(settings)

        # Format
        fmt_row = QHBoxLayout()
        fmt_row.addWidget(QLabel(self._lang.get("export_format", "Format:")))
        self._fmt_combo = QComboBox()
        self._fmt_combo.addItems(["GIF", "MP4"])
        self._fmt_combo.currentTextChanged.connect(self._on_format_changed)
        fmt_row.addWidget(self._fmt_combo)
        slay.addLayout(fmt_row)

        # FPS
        fps_row = QHBoxLayout()
        fps_row.addWidget(QLabel(self._lang.get("gif_video_fps", "FPS:")))
        self._fps_spin = QSpinBox()
        self._fps_spin.setRange(1, 60)
        self._fps_spin.setValue(5)
        fps_row.addWidget(self._fps_spin)
        fps_row.addStretch()
        slay.addLayout(fps_row)

        # Size
        size_row = QHBoxLayout()
        size_row.addWidget(QLabel(self._lang.get("gif_video_width", "Width:")))
        self._width_spin = QSpinBox()
        self._width_spin.setRange(0, 99999)
        self._width_spin.setValue(0)
        self._width_spin.setSpecialValueText(self._lang.get("gif_video_auto", "Auto"))
        size_row.addWidget(self._width_spin)
        size_row.addWidget(QLabel(self._lang.get("gif_video_height", "Height:")))
        self._height_spin = QSpinBox()
        self._height_spin.setRange(0, 99999)
        self._height_spin.setValue(0)
        self._height_spin.setSpecialValueText(self._lang.get("gif_video_auto", "Auto"))
        size_row.addWidget(self._height_spin)
        slay.addLayout(size_row)

        # Loop (GIF only)
        self._loop_check = QCheckBox(self._lang.get("gif_video_loop", "Loop forever"))
        self._loop_check.setChecked(True)
        slay.addWidget(self._loop_check)

        layout.addWidget(settings)

        # Output path
        path_row = QHBoxLayout()
        self._path_edit = QLineEdit()
        if self._paths:
            default = Path(self._paths[0]).parent / "output.gif"
            self._path_edit.setText(str(default))
        browse_btn = QPushButton(self._lang.get("export_browse", "Browse..."))
        browse_btn.clicked.connect(self._browse)
        path_row.addWidget(self._path_edit, 1)
        path_row.addWidget(browse_btn)
        layout.addLayout(path_row)

        # Progress
        self._progress = QProgressBar()
        self._progress.setVisible(False)
        layout.addWidget(self._progress)
        self._status_label = QLabel("")
        layout.addWidget(self._status_label)

        # Buttons
        btn_row = QHBoxLayout()
        btn_row.addStretch()
        cancel_btn = QPushButton(self._lang.get("export_cancel", "Cancel"))
        cancel_btn.clicked.connect(self.reject)
        self._create_btn = QPushButton(self._lang.get("gif_video_create", "Create"))
        self._create_btn.clicked.connect(self._do_create)
        btn_row.addWidget(cancel_btn)
        btn_row.addWidget(self._create_btn)
        layout.addLayout(btn_row)

    def _on_format_changed(self, text):
        is_gif = text == "GIF"
        self._loop_check.setVisible(is_gif)
        # Update extension in path
        path = self._path_edit.text()
        if path:
            p = Path(path)
            ext = ".gif" if is_gif else ".mp4"
            self._path_edit.setText(str(p.with_suffix(ext)))

    def _move_up(self):
        row = self._list.currentRow()
        if row > 0:
            item = self._list.takeItem(row)
            self._list.insertItem(row - 1, item)
            self._list.setCurrentRow(row - 1)

    def _move_down(self):
        row = self._list.currentRow()
        if row < self._list.count() - 1:
            item = self._list.takeItem(row)
            self._list.insertItem(row + 1, item)
            self._list.setCurrentRow(row + 1)

    def _browse(self):
        fmt = self._fmt_combo.currentText()
        ext = ".gif" if fmt == "GIF" else ".mp4"
        path, _ = QFileDialog.getSaveFileName(
            self, self._lang.get("gif_video_save", "Save As"),
            self._path_edit.text(),
            f"{fmt} (*{ext})",
        )
        if path:
            self._path_edit.setText(path)

    def _get_ordered_paths(self) -> list[str]:
        paths = []
        for i in range(self._list.count()):
            paths.append(self._list.item(i).data(Qt.ItemDataRole.UserRole))
        return paths

    def _do_create(self):
        output = self._path_edit.text().strip()
        if not output:
            return

        paths = self._get_ordered_paths()
        if not paths:
            return

        self._create_btn.setEnabled(False)
        self._progress.setVisible(True)
        self._progress.setMaximum(len(paths))
        self._progress.setValue(0)

        self._worker = _CreateWorker(
            paths, output,
            self._fmt_combo.currentText(),
            self._fps_spin.value(),
            self._width_spin.value(),
            self._height_spin.value(),
            self._loop_check.isChecked(),
        )
        self._worker.progress.connect(self._on_progress)
        self._worker.finished.connect(self._on_finished)
        self._worker.start()

    def _on_progress(self, current, total):
        self._progress.setValue(current)
        self._status_label.setText(f"{current}/{total}")

    def _on_finished(self, success, message):
        self._progress.setVisible(False)
        self._create_btn.setEnabled(True)
        self._worker = None

        if success:
            msg = self._lang.get("gif_video_done", "Created: {path}").format(path=message)
            self._status_label.setText(msg)
            if hasattr(self._gui.main_window, "toast"):
                self._gui.main_window.toast.success(msg)
            self.accept()
        else:
            self._status_label.setText(f"Error: {message}")
            if hasattr(self._gui.main_window, "toast"):
                self._gui.main_window.toast.info(f"Error: {message}")


def open_gif_video_dialog(main_gui: GPUImageView):
    paths = list(main_gui.selected_tiles)
    if not paths:
        return
    dlg = GifVideoDialog(main_gui, paths)
    dlg.exec()



import json
from pathlib import Path
from threading import Lock
from typing import Dict, Any, Optional, Union

from Imervue.system.app_paths import user_settings_path as _user_settings_path

# 使用者設定的全域字典
# Global dictionary for user settings
user_setting_dict: Dict[str, Any] = {
    "language": "English",
    "user_recent_folders": [],
    "user_recent_images": [],
    "user_last_folder": "",
    "bookmarks": [],
}

_lock = Lock()

def write_user_setting() -> Path:
    """
    將使用者設定寫入 JSON 檔案
    Write user settings into JSON file

    :return: 設定檔路徑 (Path to the settings file)
    """
    user_setting_file = _user_settings_path()
    write_json(str(user_setting_file), user_setting_dict)
    return user_setting_file


def read_user_setting() -> Path:
    """
    讀取使用者設定檔，並更新全域字典
    Read user settings from JSON file and update global dictionary

    :return: 設定檔路徑 (Path to the settings file)
    """
    user_setting_file = _user_settings_path()
    if user_setting_file.exists() and user_setting_file.is_file():
        data = read_json(str(user_setting_file))
        if isinstance(data, dict):
            user_setting_dict.update(data)
    return user_setting_file

def read_json(json_file_path: str) -> Optional[Any]:
    """
    use to read action file
    :param json_file_path JSON file's path to read
    """
    _lock.acquire()
    try:
        file_path = Path(json_file_path)
        if file_path.exists() and file_path.is_file():
            with open(json_file_path, encoding="utf-8") as read_file:
                return json.loads(read_file.read())
    except Exception:
        pass
    finally:
        _lock.release()


def write_json(json_save_path: str, data_to_output: Union[dict, list]) -> None:
    """
    use to save action file
    :param json_save_path  JSON save path
    :param data_to_output JSON data to output
    """
    _lock.acquire()
    try:
        # 確保父目錄存在
        Path(json_save_path).parent.mkdir(parents=True, exist_ok=True)
        with open(json_save_path, "w", encoding="utf-8") as file_to_write:
            file_to_write.write(json.dumps(data_to_output, indent=4, ensure_ascii=False))
    except Exception as e:
        import logging
        logging.getLogger("Imervue.settings").error(f"Failed to write {json_save_path}: {e}")
    finally:
        _lock.release()


traditional_chinese_word_dict = {
    "main_window_current_filename_format": "目前檔名： {name}",
    # File menu
    "main_window_open_image": "開啟檔案",
    "main_window_current_filename": "目前檔名：",
    "main_window_current_file": "檔案",
    "main_window_open_folder": "開啟資料夾",
    "main_window_exit": "離開",
    "main_window_tile_size": "縮圖大小",
    "main_window_select_folder": "選擇資料夾",
    "main_window_select_image": "選擇圖片",
    "main_window_current_folder_format": "目前資料夾： {path}",
    "main_window_remove_undo_stack": "刪除暫存撤銷檔案",
    # Language Menubar
    "menu_bar_language": "語言",
    "language_menu_bar_please_restart_messagebox": "請重新啟動",
    "language_menu_bar_english": "英文",
    "language_menu_bar_traditional_chinese": "繁體中文",
    "language_menu_bar_chinese": "簡體中文",
    "language_menu_bar_koren": "韓文",
    "language_menu_bar_japanese": "日文",
    # Tip menu
    "main_window_tip_menu": "使用說明",
    "main_window_mouse_tip_menu": "滑鼠操作",
    "main_window_keyboard_tip_menu": "鍵盤操作",
    "mouse_control_middle_tip": "可以使用滑鼠中鍵移動",
    "mouse_control_left_tip": "滑鼠左鍵點擊圖片可以進入 Deep zoom 模式",
    "mouse_control_multi_select_tip": "滑鼠左鍵按住可以在縮圖模式下複選圖片",
    "keyboard_control_esc_tip": "Deep zoom 模式下可以按 ESC 退出",
    "keyboard_control_tile_arrow_up_tip": "縮圖模式下可以使用方向鍵的上來往上移動",
    "keyboard_control_tile_arrow_down_tip": "縮圖模式下可以使用方向鍵的下來往下移動",
    "keyboard_control_tile_arrow_left_tip": "縮圖模式下可以使用方向鍵的左來往左移動",
    "keyboard_control_tile_arrow_right_tip": "縮圖模式下可以使用方向鍵的右來往右移動",
    "keyboard_control_delete_tip": "按 Delete 將圖片移至資源回收桶（縮圖和 Deep zoom 模式皆可）",
    "keyboard_r_tip": "Deep zoom 模式下按 R / Shift+R 順時針/逆時針旋轉",
    "keyboard_home_tip": "按 Home 重置縮放與位置至預設值",
    "keyboard_e_tip": "按 E 開啟圖片編輯器",
    "keyboard_undo_tip": "復原上一步操作",
    "keyboard_redo_tip": "重做上一步被復原的操作",
    "keyboard_arrow_lr_tip": "Deep zoom 模式下按左/右鍵切換圖片",
    "keyboard_arrow_tile_tip": "縮圖模式下使用方向鍵瀏覽圖片",
    "tip_mouse_middle": "中鍵",
    "tip_mouse_left_click": "左鍵點擊",
    "tip_mouse_left_drag": "左鍵拖曳",
    "tip_mouse_right_desc": "開啟右鍵選單，提供多種操作功能",
    "tip_mouse_scroll_desc": "滾輪縮放或捲動縮圖",
    "tip_close": "關閉",
    "tip_show_shortcuts": "鍵盤與滑鼠快捷鍵",
    "keyboard_f_tip": "按 F 切換全螢幕，Esc 退出全螢幕",
    "keyboard_ctrl_c_tip": "按 Ctrl+C 複製目前圖片到剪貼簿",
    "keyboard_rating_tip": "按 1~5 為目前圖片評分；再按一次取消",
    "keyboard_search_tip": "按 / 或 Ctrl+F 搜尋圖片名稱",
    "keyboard_favorite_tip": "按 0 切換目前圖片的愛心收藏",
    "keyboard_slideshow_tip": "按 S 開始/停止幻燈片播放",
    # 評分狀態
    "rating_set": "評分：{star}",
    "rating_cleared": "已清除評分",
    # 搜尋對話框
    "search_dialog_title": "搜尋圖片",
    "search_dialog_placeholder": "輸入關鍵字搜尋...",
    "search_open": "開啟",
    # Right click menu
    "right_click_menu_go_to_parent_folder": "前往父資料夾",
    "right_click_menu_next_image": "下一張圖",
    "right_click_menu_previous_image": "上一張圖",
    "right_click_menu_delete_current": "刪除目前的圖",
    "right_click_menu_delete_selected": "刪除選取的圖",
    "right_click_menu_image_info": "圖片資訊",
    # Image info
    "image_info_filename": "檔名: {info}\n",
    "image_info_fullpath": "完整路徑: {full_path}\n",
    "image_info_image_size": "尺寸: {width} x {height}\n",
    "image_info_file_size": "檔案大小: {file_size_mb} MB\n",
    "image_info_file_created_time": "建立時間: {created_time}\n",
    "image_info_file_modified_time": "修改時間: {modified_time}\n",
    "image_info_messagebox_title": "圖片資訊",
    # Image info exif
    "image_info_exif_datatime_original": "拍攝時間: {DateTimeOriginal}\n",
    "image_info_exif_camera_model": "相機: {Make} {Model}\n",
    "image_info_exif_camera_lens_model": "鏡頭: {LensModel}\n",
    "image_info_exif_camera_focal_length": "焦段: {FocalLength}\n",
    "image_info_exif_camera_fnumber": "光圈: {FNumber}\n",
    "image_info_exif_exposure_time": "快門: {ExposureTime}\n",
    "image_info_exif_iso": "ISO: {ISOSpeedRatings}",
    # Recent menu
    "recent_menu_title": "近期開啟選單",
    # Plugin
    "plugin_menu_title": "插件",
    "plugin_menu_loaded": "已載入的插件",
    "plugin_menu_no_plugins": "沒有載入任何插件",
    "plugin_menu_reload": "重新載入插件",
    "plugin_menu_open_folder": "開啟插件資料夾",
    "plugin_info_name": "名稱：{name}",
    "plugin_info_version": "版本：{version}",
    "plugin_info_author": "作者：{author}",
    "plugin_info_description": "說明：{description}",
    "plugin_info_languages": "支援語言：{languages}",
    "plugin_menu_manage": "管理插件",
    "plugin_manage_count": "已載入 {count} 個插件",
    "plugin_manage_select_hint": "選擇插件以檢視詳細資訊",
    "plugin_reload_confirm": "重新載入所有插件？這將卸載目前的插件並重新搜尋。",
    "plugin_reload_done": "已重新載入 {count} 個插件",
    "plugin_menu_download": "下載插件",
    # Plugin Downloader
    "plugin_dl_title": "下載插件",
    "plugin_dl_refresh": "重新整理",
    "plugin_dl_col_name": "插件",
    "plugin_dl_col_status": "狀態",
    "plugin_dl_download": "下載",
    "plugin_dl_reinstall": "重新安裝",
    "plugin_dl_delete": "刪除",
    "plugin_dl_close": "關閉",
    "plugin_dl_installed": "已安裝",
    "plugin_dl_not_installed": "未安裝",
    "plugin_dl_status_ready": "就緒",
    "plugin_dl_status_fetching": "正在取得插件列表...",
    "plugin_dl_status_found": "找到 {count} 個插件",
    "plugin_dl_status_error": "錯誤：{error}",
    "plugin_dl_status_downloading": "正在下載 {name}...",
    "plugin_dl_status_done": "已成功下載 {name}！",
    "plugin_dl_status_deleted": "已刪除 {name}。",
    "plugin_dl_restart_hint": "插件 '{name}' 已下載。請重新啟動 Imervue 以啟用。",
    "plugin_dl_delete_confirm_title": "刪除插件",
    "plugin_dl_delete_confirm": "確定要刪除插件 '{name}' 嗎？",
    # 右鍵選單
    "right_click_show_in_explorer": "在檔案總管中顯示",
    "right_click_copy_path": "複製路徑",
    "right_click_copy_image": "複製圖片",
    "right_click_rotate_cw": "順時針旋轉",
    "right_click_rotate_ccw": "逆時針旋轉",
    "right_click_set_wallpaper": "設為桌布",
    "right_click_compare": "比較圖片",
    "right_click_slideshow": "幻燈片播放",
    "right_click_slideshow_stop": "停止幻燈片",
    # 排序選單
    "sort_menu_title": "排序",
    "sort_by_name": "依檔名",
    "sort_by_modified": "依修改日期",
    "sort_by_created": "依建立日期",
    "sort_by_size": "依檔案大小",
    "sort_by_resolution": "依解析度",
    "sort_ascending": "升序",
    "sort_descending": "降序",
    # 篩選選單
    "filter_menu_title": "篩選",
    "filter_by_extension": "依副檔名",
    "filter_by_rating": "依評分",
    "filter_ext_all": "全部",
    "filter_ext_jpg": "JPEG",
    "filter_ext_png": "PNG",
    "filter_ext_bmp": "BMP",
    "filter_ext_tiff": "TIFF",
    "filter_ext_raw": "RAW",
    "filter_rating_all": "全部",
    "filter_rating_favorited": "已收藏",
    "filter_clear": "清除篩選",
    # 收藏
    "favorite_added": "\u2764 已收藏",
    "favorite_removed": "已取消收藏",
    # 比較
    "compare_title": "比較圖片",
    "compare_2": "比較 2 張",
    "compare_4": "比較 4 張",
    # EXIF 側邊欄
    "exif_filename": "檔案",
    "exif_filesize": "大小",
    "exif_created": "建立時間",
    "exif_modified": "修改時間",
    "exif_taken": "拍攝時間",
    "exif_make": "製造商",
    "exif_model": "型號",
    "exif_lens": "鏡頭",
    "exif_focal": "焦距",
    "exif_aperture": "光圈",
    "exif_shutter": "快門",
    "exif_iso": "ISO",
    "exif_ev": "曝光補償",
    "exif_wb": "白平衡",
    "exif_flash": "閃光燈",
    "exif_resolution": "解析度",
    "exif_no_data": "無 EXIF 資料",
    # 幻燈片
    "slideshow_title": "幻燈片播放",
    "slideshow_interval": "間隔（秒）：",
    "slideshow_fade": "淡入淡出過場",
    "slideshow_start": "開始",
    "slideshow_cancel": "取消",
    # 狀態列
    "status_loading_progress": "載入中 {current}/{total}...",
    "status_loading_image": "載入圖片中...",
    "status_ready": "就緒",
    # Toast
    "toast_path_copied": "已複製路徑",
    # 右鍵額外
    "right_click_edit_image": "編輯圖片",
    "right_click_batch": "批次操作",
    "batch_rename_title": "批次重新命名",
    "batch_rename_hint": "模板：{name} = 原始名稱，{n} = 序號，{ext} = 副檔名",
    "batch_rename_start": "起始編號：",
    "batch_rename_apply": "重新命名",
    "batch_rename_cancel": "取消",
    "batch_move_title": "移動 / 複製",
    "batch_move_mode": "模式",
    "batch_move_move": "移動",
    "batch_move_copy": "複製",
    "batch_move_apply": "套用",
    "batch_move_cancel": "取消",
    "batch_rotate_cw": "全部順時針旋轉",
    "batch_rotate_ccw": "全部逆時針旋轉",
    # 圖片編輯器
    "editor_title": "圖片編輯器",
    "editor_brightness": "亮度",
    "editor_contrast": "對比",
    "editor_saturation": "飽和度",
    "editor_apply_crop": "套用裁剪",
    "editor_rotate_cw": "順時針旋轉",
    "editor_rotate_ccw": "逆時針旋轉",
    "editor_reset": "重置",
    "editor_save_as": "另存為...",
    "editor_save": "儲存",
    "editor_saved": "已儲存！",
    # EXIF 編輯器
    "exif_edit_button": "編輯 EXIF",
    "exif_editor_title": "編輯 EXIF 元資料",
    "exif_editor_no_piexif": "需要 piexif 套件。\n安裝：pip install piexif",
    "exif_editor_fields": "元資料欄位",
    "exif_editor_save": "儲存",
    "exif_editor_cancel": "取消",
    "exif_editor_close": "關閉",
    "exif_editor_saved": "EXIF 已儲存！",
    "exif_edit_description": "描述",
    "exif_edit_artist": "作者",
    "exif_edit_copyright": "版權",
    "exif_edit_make": "製造商",
    "exif_edit_model": "型號",
    "exif_edit_user_comment": "使用者備註",
    # 匯出對話框
    "export_title": "匯出圖片",
    "export_format": "格式：",
    "export_quality": "品質：",
    "export_browse": "瀏覽...",
    "export_save": "儲存",
    "export_cancel": "取消",
    "export_saved": "匯出成功！",
    # 直方圖
    "histogram_toggle": "按 H 切換 RGB 直方圖",
    # 無損旋轉
    "lossless_rotate_cw": "無損順時針旋轉",
    "lossless_rotate_ccw": "無損逆時針旋轉",
    "lossless_rotate_done": "無損旋轉已儲存！",
    "lossless_rotate_fail": "無損旋轉失敗",
    # 右鍵選單新增
    "right_click_export": "匯出 / 另存為...",
    "right_click_lossless_rotate": "無損旋轉",
    # 動畫播放
    "anim_play": "播放",
    "anim_pause": "暫停",
    "anim_frame_prev": "上一幀",
    "anim_frame_next": "下一幀",
    "anim_frame_indicator": "幀 {current}/{total}",
    "anim_speed": "速度: {speed}x",
    # 檔案關聯
    "file_assoc_menu": "檔案關聯",
    "file_assoc_register": "註冊「以 Imervue 開啟」",
    "file_assoc_unregister": "移除檔案關聯",
    "file_assoc_done": "檔案關聯已註冊！",
    "file_assoc_removed": "檔案關聯已移除！",
    "file_assoc_need_admin": "需要管理員權限",
    # 多視窗
    "menu_new_window": "新視窗",
    # 適應模式
    "fit_width": "適應寬度",
    "fit_height": "適應高度",
    "keyboard_w_tip": "按 W 適應寬度，Shift+W 適應高度",
    # 額外快捷鍵提示
    "keyboard_h_tip": "在深度縮放模式下按 H 切換 RGB 直方圖",
    "keyboard_ctrl_v_tip": "按 Ctrl+V 從剪貼簿貼上圖片",
    "keyboard_anim_play_tip": "按空白鍵播放/暫停動畫圖片（GIF/APNG/WebP）",
    "keyboard_anim_frame_tip": "按 < / > 逐幀瀏覽動畫",
    "keyboard_anim_speed_tip": "按 [ / ] 減慢/加快動畫播放速度",
    # 書籤
    "bookmark_title": "書籤",
    "bookmark_open": "開啟",
    "bookmark_remove": "移除",
    "bookmark_clear": "清空全部",
    "bookmark_close": "關閉",
    "bookmark_count": "{count} 張已收藏圖片",
    "bookmark_added": "\u2605 已收藏",
    "bookmark_removed": "已取消收藏",
    "bookmark_add_current": "新增收藏",
    "bookmark_remove_current": "取消收藏",
    "bookmark_clear_confirm_title": "清空書籤",
    "bookmark_clear_confirm": "確定清空所有書籤？",
    "keyboard_b_tip": "按 B 切換當前圖片的收藏狀態",
    # SVG
    "filter_ext_svg": "SVG",
    # 標籤與相簿
    "tag_album_title": "標籤與相簿",
    "tag_tab_tags": "標籤",
    "tag_tab_albums": "相簿",
    "tag_close": "關閉",
    "tag_count": "{count} 個標籤",
    "tag_create": "新增標籤",
    "tag_rename": "重新命名",
    "tag_delete": "刪除",
    "tag_images": "標籤中的圖片：",
    "tag_remove_image": "從標籤移除",
    "tag_filter": "在檢視器中顯示",
    "tag_create_prompt": "標籤名稱：",
    "tag_rename_prompt": "新名稱：",
    "tag_delete_confirm_title": "刪除標籤",
    "tag_delete_confirm": "確定刪除標籤「{name}」？",
    "tag_menu_title": "標籤",
    "tag_create_and_add": "+ 新增標籤...",
    "album_count": "{count} 個相簿",
    "album_create": "新增相簿",
    "album_rename": "重新命名",
    "album_delete": "刪除",
    "album_images": "相簿中的圖片：",
    "album_remove_image": "從相簿移除",
    "album_view": "瀏覽相簿",
    "album_create_prompt": "相簿名稱：",
    "album_rename_prompt": "新名稱：",
    "album_delete_confirm_title": "刪除相簿",
    "album_delete_confirm": "確定刪除相簿「{name}」？",
    "album_menu_title": "相簿",
    "album_create_and_add": "+ 新增相簿...",
    "batch_tag_title": "加入標籤",
    "batch_album_title": "加入相簿",
    "filter_by_tag": "依標籤",
    "filter_by_album": "依相簿",
    "filter_no_tags": "（無標籤）",
    "filter_no_albums": "（無相簿）",
    "keyboard_t_tip": "按 T 開啟標籤與相簿管理",
    # 批次匯出
    "batch_export_title": "批次匯出",
    "batch_export_count": "已選取 {count} 張圖片",
    "batch_export_resize": "調整大小",
    "batch_export_max_width": "最大寬度：",
    "batch_export_max_height": "最大高度：",
    "batch_export_start": "匯出",
    "batch_export_done": "已匯出 {success}/{total} 張圖片",
    # GIF / 影片
    "gif_video_title": "製作 GIF / 影片",
    "gif_video_order": "拖曳排序（最上方 = 第一幀）：",
    "gif_video_move_up": "上移",
    "gif_video_move_down": "下移",
    "gif_video_settings": "設定",
    "gif_video_fps": "幀率：",
    "gif_video_width": "寬度：",
    "gif_video_height": "高度：",
    "gif_video_auto": "自動",
    "gif_video_loop": "無限循環",
    "gif_video_create": "建立",
    "gif_video_save": "另存為",
    "gif_video_done": "已建立：{path}",
}